# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import List, Optional

from gravitino.api.function.function import Function
from gravitino.api.function.function_catalog import FunctionCatalog
from gravitino.api.function.function_change import (
    AddDefinition,
    AddImpl,
    FunctionChange,
    RemoveDefinition,
    RemoveImpl,
    UpdateComment,
    UpdateImpl,
)
from gravitino.api.function.function_definition import FunctionDefinition
from gravitino.api.function.function_type import FunctionType
from gravitino.client.generic_function import GenericFunction
from gravitino.dto.function.function_definition_dto import FunctionDefinitionDTO
from gravitino.dto.function.function_impl_dto import (
    function_impl_dto_from_function_impl,
)
from gravitino.dto.function.function_param_dto import FunctionParamDTO
from gravitino.dto.requests.function_register_request import FunctionRegisterRequest
from gravitino.dto.requests.function_update_request import (
    AddDefinitionRequest,
    AddImplRequest,
    FunctionUpdateRequest,
    RemoveDefinitionRequest,
    RemoveImplRequest,
    UpdateCommentRequest,
    UpdateImplRequest,
)
from gravitino.dto.requests.function_updates_request import FunctionUpdatesRequest
from gravitino.dto.responses.drop_response import DropResponse
from gravitino.dto.responses.entity_list_response import EntityListResponse
from gravitino.dto.responses.function_list_response import FunctionListResponse
from gravitino.dto.responses.function_response import FunctionResponse
from gravitino.exceptions.base import IllegalArgumentException
from gravitino.exceptions.handlers.function_error_handler import (
    FUNCTION_ERROR_HANDLER,
)
from gravitino.name_identifier import NameIdentifier
from gravitino.namespace import Namespace
from gravitino.rest import rest_utils


class FunctionCatalogOperations(FunctionCatalog):
    """Function catalog operations helper class that provides implementations for function management.

    This class is used by catalogs that support function operations.
    """

    def __init__(self, rest_client, catalog_namespace: Namespace, catalog_name: str):
        """Create a FunctionCatalogOperations instance.

        Args:
            rest_client: The REST client for making API calls.
            catalog_namespace: The namespace of the catalog.
            catalog_name: The name of the catalog.
        """
        self._rest_client = rest_client
        self._catalog_namespace = catalog_namespace
        self._catalog_name = catalog_name

    def list_functions(self, namespace: Namespace) -> List[NameIdentifier]:
        """List the functions in a schema namespace from the catalog.

        Args:
            namespace: A schema namespace. This namespace should have 1 level,
                       which is the schema name.

        Returns:
            A list of function identifiers in the namespace.

        Raises:
            NoSuchSchemaException: If the schema does not exist.
        """
        self._check_function_namespace(namespace)

        full_namespace = self._get_function_full_namespace(namespace)
        resp = self._rest_client.get(
            self._format_function_request_path(full_namespace),
            error_handler=FUNCTION_ERROR_HANDLER,
        )

        entity_list_response = EntityListResponse.from_json(resp.body)
        entity_list_response.validate()

        return [
            NameIdentifier.of(ident.namespace().level(2), ident.name())
            for ident in entity_list_response.identifiers()
        ]

    def list_function_infos(self, namespace: Namespace) -> List[Function]:
        """List the functions with details in a schema namespace from the catalog.

        Args:
            namespace: A namespace.

        Returns:
            A list of functions in the namespace.

        Raises:
            NoSuchSchemaException: If the schema does not exist.
        """
        self._check_function_namespace(namespace)

        full_namespace = self._get_function_full_namespace(namespace)
        params = {"details": "true"}
        resp = self._rest_client.get(
            self._format_function_request_path(full_namespace),
            params=params,
            error_handler=FUNCTION_ERROR_HANDLER,
        )

        function_list_response = FunctionListResponse.from_json(resp.body)
        function_list_response.validate()

        return [GenericFunction(func) for func in function_list_response.functions()]

    def get_function(self, ident: NameIdentifier) -> Function:
        """Get a function by NameIdentifier from the catalog.

        Args:
            ident: A function identifier, which should be "schema.function" format.

        Returns:
            The function metadata.

        Raises:
            NoSuchFunctionException: If the function does not exist.
        """
        self._check_function_name_identifier(ident)

        full_namespace = self._get_function_full_namespace(ident.namespace())
        resp = self._rest_client.get(
            f"{self._format_function_request_path(full_namespace)}/"
            f"{rest_utils.encode_string(ident.name())}",
            error_handler=FUNCTION_ERROR_HANDLER,
        )

        function_response = FunctionResponse.from_json(resp.body)
        function_response.validate()

        return GenericFunction(function_response.function())

    def register_function(
        self,
        ident: NameIdentifier,
        comment: Optional[str],
        function_type: FunctionType,
        deterministic: bool,
        definitions: List[FunctionDefinition],
    ) -> Function:
        """Register a function with one or more definitions (overloads).

        Args:
            ident: The function identifier.
            comment: The optional function comment.
            function_type: The function type.
            deterministic: Whether the function is deterministic.
            definitions: The function definitions.

        Returns:
            The registered function.

        Raises:
            NoSuchSchemaException: If the schema does not exist.
            FunctionAlreadyExistsException: If the function already exists.
        """
        self._check_function_name_identifier(ident)

        full_namespace = self._get_function_full_namespace(ident.namespace())

        # Convert definitions to DTOs
        definition_dtos = [
            FunctionDefinitionDTO.from_function_definition(d) for d in definitions
        ]

        req = FunctionRegisterRequest(
            name=ident.name(),
            function_type=function_type,
            deterministic=deterministic,
            definitions=definition_dtos,
            comment=comment,
        )
        req.validate()

        resp = self._rest_client.post(
            self._format_function_request_path(full_namespace),
            req,
            error_handler=FUNCTION_ERROR_HANDLER,
        )

        function_response = FunctionResponse.from_json(resp.body)
        function_response.validate()

        return GenericFunction(function_response.function())

    def alter_function(
        self, ident: NameIdentifier, *changes: FunctionChange
    ) -> Function:
        """Applies FunctionChange changes to a function in the catalog.

        Args:
            ident: The NameIdentifier instance of the function to alter.
            changes: The FunctionChange instances to apply to the function.

        Returns:
            The updated Function instance.

        Raises:
            NoSuchFunctionException: If the function does not exist.
            IllegalArgumentException: If the change is rejected by the implementation.
        """
        self._check_function_name_identifier(ident)
        if not changes:
            raise IllegalArgumentException(
                "At least one FunctionChange must be provided to alter_function"
            )

        full_namespace = self._get_function_full_namespace(ident.namespace())
        updates = [self._to_function_update_request(change) for change in changes]
        req = FunctionUpdatesRequest(updates)
        req.validate()

        resp = self._rest_client.put(
            f"{self._format_function_request_path(full_namespace)}/"
            f"{rest_utils.encode_string(ident.name())}",
            req,
            error_handler=FUNCTION_ERROR_HANDLER,
        )

        function_response = FunctionResponse.from_json(resp.body)
        function_response.validate()

        return GenericFunction(function_response.function())

    def drop_function(self, ident: NameIdentifier) -> bool:
        """Drop a function from the catalog.

        Args:
            ident: A function identifier, which should be "schema.function" format.

        Returns:
            True if the function is dropped, False if the function did not exist.
        """
        self._check_function_name_identifier(ident)

        full_namespace = self._get_function_full_namespace(ident.namespace())
        resp = self._rest_client.delete(
            f"{self._format_function_request_path(full_namespace)}/"
            f"{rest_utils.encode_string(ident.name())}",
            error_handler=FUNCTION_ERROR_HANDLER,
        )

        drop_response = DropResponse.from_json(resp.body)
        drop_response.validate()

        return drop_response.dropped()

    def _format_function_request_path(self, ns: Namespace) -> str:
        """Format the request path for function operations.

        Args:
            ns: The full namespace (metalake.catalog.schema).

        Returns:
            The request path.
        """
        return (
            f"api/metalakes/{rest_utils.encode_string(ns.level(0))}"
            f"/catalogs/{rest_utils.encode_string(ns.level(1))}"
            f"/schemas/{rest_utils.encode_string(ns.level(2))}"
            f"/functions"
        )

    def _check_function_namespace(self, namespace: Namespace):
        """Check whether the namespace of a function is valid.

        Args:
            namespace: The namespace to check.

        Raises:
            IllegalArgumentException: If the namespace is invalid.
        """
        if namespace is None or namespace.length() != 1:
            raise IllegalArgumentException(
                f"Function namespace must be non-null and have 1 level, "
                f"the input namespace is {namespace}"
            )

    def _check_function_name_identifier(self, ident: NameIdentifier):
        """Check whether the NameIdentifier of a function is valid.

        Args:
            ident: The NameIdentifier to check.

        Raises:
            IllegalArgumentException: If the identifier is invalid.
        """
        if ident is None:
            raise IllegalArgumentException("NameIdentifier must not be null")
        if not ident.name():
            raise IllegalArgumentException("NameIdentifier name must not be empty")
        self._check_function_namespace(ident.namespace())

    def _get_function_full_namespace(self, function_namespace: Namespace) -> Namespace:
        """Get the full namespace of the function with the given function's short namespace.

        Args:
            function_namespace: The function's short namespace (schema name).

        Returns:
            The full namespace (metalake.catalog.schema).
        """
        return Namespace.of(
            self._catalog_namespace.level(0),
            self._catalog_name,
            function_namespace.level(0),
        )

    def _to_function_update_request(
        self, change: FunctionChange
    ) -> FunctionUpdateRequest:
        """Convert a FunctionChange to a FunctionUpdateRequest.

        Args:
            change: The function change.

        Returns:
            The corresponding update request.

        Raises:
            IllegalArgumentException: If the change type is not supported.
        """
        if isinstance(change, UpdateComment):
            return UpdateCommentRequest(change.new_comment())
        if isinstance(change, AddDefinition):
            definition_dto = FunctionDefinitionDTO.from_function_definition(
                change.definition()
            )
            return AddDefinitionRequest(definition_dto)
        if isinstance(change, RemoveDefinition):
            param_dtos = [
                FunctionParamDTO.from_function_param(p) for p in change.parameters()
            ]
            return RemoveDefinitionRequest(param_dtos)
        if isinstance(change, AddImpl):
            param_dtos = [
                FunctionParamDTO.from_function_param(p) for p in change.parameters()
            ]
            impl_dto = function_impl_dto_from_function_impl(change.implementation())
            return AddImplRequest(param_dtos, impl_dto)
        if isinstance(change, UpdateImpl):
            param_dtos = [
                FunctionParamDTO.from_function_param(p) for p in change.parameters()
            ]
            impl_dto = function_impl_dto_from_function_impl(change.implementation())
            return UpdateImplRequest(param_dtos, change.runtime().name, impl_dto)
        if isinstance(change, RemoveImpl):
            param_dtos = [
                FunctionParamDTO.from_function_param(p) for p in change.parameters()
            ]
            return RemoveImplRequest(param_dtos, change.runtime().name)
        raise IllegalArgumentException(f"Unknown function change type: {type(change)}")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass, field
from dataclasses_json import config, DataClassJsonMixin

from gravitino.api.job.job_handle import JobHandle
from gravitino.dto.audit_dto import AuditDTO


@dataclass
class JobDTO(DataClassJsonMixin):
    """Data transfer object representing a Job."""

    _job_id: str = field(metadata=config(field_name="jobId"))
    _job_template_name: str = field(metadata=config(field_name="jobTemplateName"))
    _status: JobHandle.Status = field(
        metadata=config(
            field_name="status",
            encoder=JobHandle.Status.job_status_serialize,
            decoder=JobHandle.Status.job_status_deserialize,
        )
    )
    _audit: AuditDTO = field(metadata=config(field_name="audit"))

    def job_id(self) -> str:
        """Returns the job ID."""
        return self._job_id

    def job_template_name(self) -> str:
        """Returns the job template name."""
        return self._job_template_name

    def status(self) -> JobHandle.Status:
        """Returns the status of the job."""
        return self._status

    def audit(self) -> AuditDTO:
        """Returns the audit information of the job."""
        return self._audit

    def validate(self) -> None:
        """Validates the JobDTO, ensuring required fields are present and non-empty."""
        if self._job_id is None or not self._job_id.strip():
            raise ValueError('"jobId" is required and cannot be empty')
        if self._job_template_name is None or not self._job_template_name.strip():
            raise ValueError('"jobTemplateName" is required and cannot be empty')
        if self._status is None:
            raise ValueError('"status" must not be None')
        if self._audit is None:
            raise ValueError('"audit" must not be None')



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import json
from dataclasses import dataclass, field
from typing import List

from dataclasses_json import config
from gravitino.exceptions.base import IllegalArgumentException

from .base_response import BaseResponse
from ..job.job_template_dto import JobTemplateDTO


@dataclass
class JobTemplateListResponse(BaseResponse):
    """Represents a response containing a list of job templates."""

    _job_templates: List[JobTemplateDTO] = field(
        metadata=config(field_name="jobTemplates")
    )

    def validate(self):
        """Validates the response data.

        Raises:
            IllegalArgumentException: If the job templates list is not set or contains invalid items.
        """
        super().validate()

        if self._job_templates is None:
            raise IllegalArgumentException("jobTemplates must not be None")

        for job_template in self._job_templates:
            if job_template is not None:
                job_template.validate()

    def job_templates(self) -> List[JobTemplateDTO]:
        """Returns the list of job templates from the response."""
        return self._job_templates

    @classmethod
    def from_json(
        cls, s: str, infer_missing: bool = False, **kwargs
    ) -> "JobTemplateListResponse":
        """Deserialize JSON string into a JobTemplateListResponse object."""

        data = json.loads(s)

        # Deserialize each job template using JobTemplateDTO.from_json
        job_templates = [
            JobTemplateDTO.from_json(
                json.dumps(item), infer_missing=infer_missing, **kwargs
            )
            for item in data.get("jobTemplates", [])
        ]

        return cls(_job_templates=job_templates, _code=data["code"])



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass, field

from dataclasses_json import config

from gravitino.dto.rel.table_dto import TableDTO
from gravitino.dto.responses.base_response import BaseResponse
from gravitino.utils.precondition import Precondition


@dataclass
class TableResponse(BaseResponse):
    """Represents a response for a table."""

    _table: TableDTO = field(metadata=config(field_name="table"))

    def table(self) -> TableDTO:
        return self._table

    def validate(self) -> None:
        Precondition.check_string_not_empty(
            self._table.name(), "table 'name' must not be null and empty"
        )
        Precondition.check_argument(
            self._table.audit_info() is not None, "table 'audit' must not be null"
        )
        Precondition.check_argument(
            self._table.partitioning() is not None,
            "table 'partitions' must not be null",
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import abstractmethod
from enum import Enum, unique
from typing import Final, List

from gravitino.api.rel.expressions.expression import Expression
from gravitino.api.rel.expressions.named_reference import NamedReference
from gravitino.api.rel.expressions.transforms.transform import Transform
from gravitino.dto.rel.partition_utils import PartitionUtils
from gravitino.exceptions.base import IllegalArgumentException
from gravitino.utils.precondition import Precondition


class Partitioning(Transform):
    EMPTY_PARTITIONING: Final[List["Partitioning"]] = []
    """An empty array of partitioning."""

    @unique
    class Strategy(str, Enum):
        """Represents a partitioning strategy."""

        IDENTITY = "identity"
        """The identity partitioning strategy."""

        YEAR = "year"
        """The year partitioning strategy."""

        MONTH = "month"
        """The month partitioning strategy."""

        DAY = "day"
        """The day partitioning strategy."""

        HOUR = "hour"
        """The hour partitioning strategy."""

        BUCKET = "bucket"
        """The bucket partitioning strategy."""

        TRUNCATE = "truncate"
        """The truncate partitioning strategy."""

        LIST = "list"
        """The list partitioning strategy."""

        RANGE = "range"
        """The range partitioning strategy."""

        FUNCTION = "function"
        """The function partitioning strategy."""

    @abstractmethod
    def strategy(self) -> Strategy:
        """Returns the name of the partitioning strategy.

        Returns:
            Strategy: The name of the partitioning strategy.
        """
        pass  # pragma: no cover

    @abstractmethod
    def validate(self, columns: List["ColumnDTO"]) -> None:
        """Validates the partitioning columns.

        Args:
            columns (List[ColumnDTO]): The columns to be validated.

        Raises:
            IllegalArgumentException: If the columns are invalid, this exception is thrown.
        """
        pass  # pragma: no cover

    @staticmethod
    def get_by_name(name: str) -> Strategy:
        """Gets the partitioning strategy by name.

        Args:
            name (str): The name of the partitioning strategy.

        Returns:
            Strategy: The partitioning strategy.

        Raises:
            IllegalArgumentException: If the name is invalid, this exception is thrown.
        """

        for strategy in Partitioning.Strategy:
            if strategy.name.lower() == name.lower():
                return strategy
        raise IllegalArgumentException(
            f"Invalid partitioning strategy: {name}. "
            f"Valid values are: {', '.join([str(s) for s in Partitioning.Strategy])}"
        )


class SingleFieldPartitioning(Partitioning):
    """A single field partitioning strategy."""

    def __init__(self, field_name: List[str]):
        Precondition.check_argument(
            field_name is not None and len(field_name) > 0,
            "field_name cannot be null or empty",
        )
        self._field_name = field_name

    def field_name(self) -> List[str]:
        """Returns the field name of the partitioning.

        Returns:
            List[str]: The field name of the partitioning.
        """
        return self._field_name

    def validate(self, columns: List["ColumnDTO"]) -> None:
        """Validates the partitioning columns.

        Args:
            columns (List[ColumnDTO]): The columns to be validated.

        Raises:
            IllegalArgumentException: If the columns are invalid, this exception is thrown.
        """

        PartitionUtils.validate_field_existence(columns, self._field_name)

    def name(self) -> str:
        """Returns the name of the partitioning strategy.

        Returns:
            str: The name of the partitioning strategy.
        """
        return self.strategy().name.lower()

    def arguments(self) -> List[Expression]:
        """Returns the arguments of the partitioning strategy.

        Returns:
            List[Expression]: The arguments of the partitioning strategy.
        """
        return [NamedReference.field(self._field_name)]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import ClassVar, List, Optional, Union, overload

from gravitino.api.rel.expressions.expression import Expression
from gravitino.api.rel.expressions.literals.literal import Literal
from gravitino.api.rel.expressions.literals.literals import Literals
from gravitino.api.rel.expressions.named_reference import NamedReference
from gravitino.api.rel.expressions.transforms.transform import (
    SingleFieldTransform,
    Transform,
)
from gravitino.api.rel.partitions.list_partition import ListPartition
from gravitino.api.rel.partitions.partition import Partition
from gravitino.api.rel.partitions.range_partition import RangePartition


class Transforms(Transform):
    """Helper methods to create logical transforms to pass into Apache Gravitino."""

    EMPTY_TRANSFORM: ClassVar[List[Transform]] = []
    """An empty array of transforms."""
    NAME_OF_IDENTITY: ClassVar[str] = "identity"
    """The name of the identity transform."""
    NAME_OF_YEAR: ClassVar[str] = "year"
    """The name of the year transform. The year transform returns the year of the input value."""
    NAME_OF_MONTH: ClassVar[str] = "month"
    """The name of the month transform. The month transform returns the month of the input value."""
    NAME_OF_DAY: ClassVar[str] = "day"
    """The name of the day transform. The day transform returns the day of the input value."""
    NAME_OF_HOUR: ClassVar[str] = "hour"
    """The name of the hour transform. The hour transform returns the hour of the input value."""
    NAME_OF_BUCKET: ClassVar[str] = "bucket"
    """The name of the bucket transform. The bucket transform returns the bucket of the input value."""
    NAME_OF_TRUNCATE: ClassVar[str] = "truncate"
    """The name of the truncate transform. The truncate transform returns the truncated value of the"""
    NAME_OF_LIST: ClassVar[str] = "list"
    """The name of the list transform. The list transform includes multiple fields in a list."""
    NAME_OF_RANGE: ClassVar[str] = "range"
    """The name of the range transform. The range transform returns the range of the input value."""

    @staticmethod
    @overload
    def identity(field_name: List[str]) -> "Transforms.IdentityTransform": ...

    @staticmethod
    @overload
    def identity(field_name: str) -> "Transforms.IdentityTransform": ...

    @staticmethod
    def identity(field_name: Union[str, List[str]]) -> "Transforms.IdentityTransform":
        """Create a transform that returns the input value.

        Args:
            field_name (List[str]):
                The field name(s) to transform. Can be a list of field names or a single field name.
        Returns:
            Transforms.IdentityTransform: The created transform
        """

        return Transforms.IdentityTransform(
            NamedReference.field(
                [field_name] if isinstance(field_name, str) else field_name
            )
        )

    @staticmethod
    @overload
    def year(field_name: List[str]) -> "Transforms.YearTransform": ...

    @staticmethod
    @overload
    def year(field_name: str) -> "Transforms.YearTransform": ...

    @staticmethod
    def year(field_name: Union[str, List[str]]) -> "Transforms.YearTransform":
        """Create a transform that returns the input value.

        Args:
            field_name (List[str]):
                The field name(s) to transform. Can be a list of field names or a single field name.
        Returns:
            Transforms.YearTransform: The created transform
        """

        return Transforms.YearTransform(
            NamedReference.field(
                [field_name] if isinstance(field_name, str) else field_name
            )
        )

    @staticmethod
    @overload
    def month(field_name: List[str]) -> "Transforms.MonthTransform": ...

    @staticmethod
    @overload
    def month(field_name: str) -> "Transforms.MonthTransform": ...

    @staticmethod
    def month(field_name: Union[str, List[str]]) -> "Transforms.MonthTransform":
        """Create a transform that returns the input value.

        Args:
            field_name (List[str]):
                The field name(s) to transform. Can be a list of field names or a single field name.
        Returns:
            MonthTransform: The created transform
        """

        return Transforms.MonthTransform(
            NamedReference.field(
                [field_name] if isinstance(field_name, str) else field_name
            )
        )

    @staticmethod
    @overload
    def day(field_name: List[str]) -> "Transforms.DayTransform": ...

    @staticmethod
    @overload
    def day(field_name: str) -> "Transforms.DayTransform": ...

    @staticmethod
    def day(field_name: Union[str, List[str]]) -> "Transforms.DayTransform":
        """Create a transform that returns the input value.

        Args:
            field_name (List[str]):
                The field name(s) to transform. Can be a list of field names or a single field name.
        Returns:
            DayTransform: The created transform
        """

        return Transforms.DayTransform(
            NamedReference.field(
                [field_name] if isinstance(field_name, str) else field_name
            )
        )

    @staticmethod
    @overload
    def hour(field_name: List[str]) -> "Transforms.HourTransform": ...

    @staticmethod
    @overload
    def hour(field_name: str) -> "Transforms.HourTransform": ...

    @staticmethod
    def hour(field_name: Union[str, List[str]]) -> "Transforms.HourTransform":
        """Create a transform that returns the input value.

        Args:
            field_name (List[str]):
                The field name(s) to transform. Can be a list of field names or a single field name.
        Returns:
            Transforms.HourTransform: The created transform
        """

        return Transforms.HourTransform(
            NamedReference.field(
                [field_name] if isinstance(field_name, str) else field_name
            )
        )

    @staticmethod
    def bucket(
        num_buckets: int, *field_names: List[str]
    ) -> "Transforms.BucketTransform":
        """Create a transform that returns the bucket of the input value.

        Args:
            num_buckets (int): The number of buckets to use
            *field_names (List[str]): The field names to transform

        Returns:
            Transforms.BucketTransform: The created transform
        """
        return Transforms.BucketTransform(
            num_buckets=Literals.integer_literal(value=num_buckets),
            fields=[NamedReference.field(names) for names in field_names],
        )

    @staticmethod
    @overload
    def truncate(
        width: int, field_name: List[str]
    ) -> "Transforms.TruncateTransform": ...

    @staticmethod
    @overload
    def truncate(width: int, field_name: str) -> "Transforms.TruncateTransform": ...

    @staticmethod
    def truncate(
        width: int, field_name: Union[str, List[str]]
    ) -> "Transforms.TruncateTransform":
        """Create a transform that returns the truncated value of the input value with the given width.

        Args:
            width (int): The width to truncate to
            field_name (Union[str, List[str]]): The column/field name to transform

        Returns:
            Transforms.TruncateTransform: The created transform
        """
        return Transforms.TruncateTransform(
            width=Literals.integer_literal(value=width),
            field=NamedReference.field(
                [field_name] if isinstance(field_name, str) else field_name
            ),
        )

    @staticmethod
    def apply(name: str, arguments: List[Expression]) -> "Transforms.ApplyTransform":
        """Create a transform that applies a function to the input value.

        Args:
            name (str): The name of the function to apply
            arguments (List[Expression]): he arguments to the function

        Returns:
            Transforms.ApplyTransform: The created transform
        """

        return Transforms.ApplyTransform(name=name, arguments=arguments)

    @overload
    @staticmethod
    def list(*field_names: List[str]) -> "Transforms.ListTransform": ...

    @overload
    @staticmethod
    def list(
        *, field_names: List[List[str]], assignments: List[ListPartition]
    ) -> "Transforms.ListTransform": ...

    @staticmethod
    def list(*args, **kwargs) -> "Transforms.ListTransform":
        """Create a transform that includes multiple fields in a list with preassigned list partitions.

        Args:
            field_names (List[str] or List[List[str]]):
                The field names to include in the list
            assignments (Optional[List[ListPartition]]):
                The preassigned list partitions

        Returns:
            Transforms.ListTransform: The created transform
        """
        field_names = kwargs.get("field_names", args)
        assignments = kwargs.get("assignments")
        return Transforms.ListTransform(
            fields=[
                NamedReference.field(field_name=field_name)
                for field_name in field_names
            ],
            assignments=[] if assignments is None else assignments,
        )

    @staticmethod
    def range(
        field_name: List[str], assignments: Optional[List[RangePartition]] = None
    ) -> "Transforms.RangeTransform":
        """Create a transform that returns the range of the input value with preassigned range partitions.

        Args:
            field_name (List[str]):
                The field name to transform
            assignments (Optional[List[RangePartition]], optional):
                The preassigned range partitions. Defaults to `None`.

        Returns:
            Transforms.RangeTransform: The created transform
        """
        return Transforms.RangeTransform(
            field=NamedReference.field(field_name=field_name),
            assignments=[] if assignments is None else assignments,
        )

    class IdentityTransform(SingleFieldTransform):
        """A transform that returns the input value."""

        def name(self) -> str:
            return Transforms.NAME_OF_IDENTITY

    class YearTransform(SingleFieldTransform):
        """A transform that returns the year of the input value."""

        def name(self) -> str:
            return Transforms.NAME_OF_YEAR

    class MonthTransform(SingleFieldTransform):
        """A transform that returns the month of the input value."""

        def name(self) -> str:
            return Transforms.NAME_OF_MONTH

    class DayTransform(SingleFieldTransform):
        """A transform that returns the day of the input value."""

        def name(self) -> str:
            return Transforms.NAME_OF_DAY

    class HourTransform(SingleFieldTransform):
        """A transform that returns the hour of the input value."""

        def name(self) -> str:
            return Transforms.NAME_OF_HOUR

    class BucketTransform(Transform):
        """A transform that returns the bucket of the input value."""

        def __init__(self, num_buckets: Literal[int], fields: List[NamedReference]):
            self.num_buckets_ = num_buckets
            self.fields = fields

        def name(self) -> str:
            return Transforms.NAME_OF_BUCKET

        def num_buckets(self) -> int:
            return self.num_buckets_.value()

        def field_names(self) -> List[List[str]]:
            return [field.field_name() for field in self.fields]

        def arguments(self) -> List[Expression]:
            return [self.num_buckets_, *self.fields]

        def __eq__(self, value: object) -> bool:
            if self is value:
                return True
            return (
                isinstance(value, Transforms.BucketTransform)
                and self.num_buckets_ == value.num_buckets_
                and self.fields == value.fields
            )

        def __hash__(self) -> int:
            return hash((self.num_buckets_, *self.fields))

    class TruncateTransform(Transform):
        """A transform that returns the truncated value of the input value with the given width."""

        def __init__(self, width: Literal[int], field: NamedReference):
            self.width_ = width
            self.field = field

        def name(self) -> str:
            return Transforms.NAME_OF_TRUNCATE

        def width(self) -> int:
            """Gets the width to truncate to.

            Returns:
                int: The width to truncate to.
            """

            return self.width_.value()

        def field_name(self) -> List[str]:
            """Gets the field name to transform.

            Returns:
                List[str]: The field name to transform.
            """

            return self.field.field_name()

        def arguments(self) -> List[Expression]:
            return [self.width_, self.field]

        def __eq__(self, value: object) -> bool:
            if self is value:
                return True
            return (
                isinstance(value, Transforms.TruncateTransform)
                and self.width_ == value.width_
                and self.field == value.field
            )

        def __hash__(self) -> int:
            return hash((self.width_, self.field))

    class ApplyTransform(Transform):
        """A transform that applies a function to the input value."""

        def __init__(self, name: str, arguments: List[Expression]):
            self.name_ = name
            self.arguments_ = arguments

        def name(self) -> str:
            return self.name_

        def arguments(self) -> List[Expression]:
            return self.arguments_

        def __eq__(self, value: object) -> bool:
            if self is value:
                return True
            return (
                isinstance(value, Transforms.ApplyTransform)
                and self.name_ == value.name_
                and self.arguments_ == value.arguments_
            )

        def __hash__(self) -> int:
            return 31 * hash(self.name_) + hash(tuple(self.arguments_))

    class ListTransform(Transform):
        """A transform that includes multiple fields in a list."""

        def __init__(
            self,
            fields: List[NamedReference],
            assignments: Optional[List[ListPartition]] = None,
        ):
            self._fields = fields
            self._assignments = [] if assignments is None else assignments

        def field_names(self) -> List[List[str]]:
            """Gets the field names to include in the list.

            Returns:
                List[List[str]]: The field names to include in the list.
            """
            return [field.field_name() for field in self._fields]

        def name(self) -> str:
            return Transforms.NAME_OF_LIST

        def arguments(self) -> List[Expression]:
            return self._fields

        def assignments(self) -> List[Partition]:
            return self._assignments

        def __eq__(self, value: object) -> bool:
            if not isinstance(value, Transforms.ListTransform):
                return False
            return self is value or self._fields == value.arguments()

        def __hash__(self) -> int:
            return hash(tuple(self._fields))

    class RangeTransform(Transform):
        """A transform that returns the range of the input value."""

        def __init__(
            self,
            field: NamedReference,
            assignments: Optional[List[RangePartition]] = None,
        ):
            self._field = field
            self._assignments = [] if assignments is None else assignments

        def name(self) -> str:
            return Transforms.NAME_OF_RANGE

        def field_name(self) -> List[str]:
            """Gets the field name to transform.

            Returns:
                List[str]: The field name to transform.
            """

            return self._field.field_name()

        def arguments(self) -> List[Expression]:
            return [self._field]

        def assignments(self) -> List[Partition]:
            return self._assignments

        def __eq__(self, value: object) -> bool:
            if not isinstance(value, Transforms.RangeTransform):
                return False
            return self is value or self.field_name() == value.field_name()

        def __hash__(self) -> int:
            return hash(self._field)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from typing import Any

from gravitino.api.rel.expressions.distributions.strategy import Strategy
from gravitino.api.rel.types.json_serdes.base import JsonSerializable
from gravitino.dto.rel.distribution_dto import DistributionDTO
from gravitino.dto.rel.expressions.json_serdes._helper.serdes_utils import SerdesUtils
from gravitino.utils.precondition import Precondition
from gravitino.utils.serdes import SerdesUtilsBase


class DistributionSerDes(SerdesUtilsBase, JsonSerializable[DistributionDTO]):
    """Custom JSON deserializer for DistributionDTO objects."""

    @classmethod
    def serialize(cls, data_type: DistributionDTO) -> dict[str, Any]:
        return {
            cls.STRATEGY: data_type.strategy().name.lower(),
            cls.NUMBER: data_type.number(),
            cls.FUNCTION_ARGS: [
                SerdesUtils.write_function_arg(arg) for arg in data_type.args()
            ],
        }

    @classmethod
    def deserialize(cls, data: dict[str, Any]) -> DistributionDTO:
        Precondition.check_argument(
            isinstance(data, dict) and len(data) > 0,
            f"Cannot parse distribution from invalid JSON: {data}",
        )
        strategy_data = data.get(cls.STRATEGY, Strategy.HASH.value)
        return DistributionDTO(
            strategy=Strategy(strategy_data.upper()),
            number=data[cls.NUMBER],
            args=[
                SerdesUtils.read_function_arg(arg) for arg in data[cls.FUNCTION_ARGS]
            ],
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from functools import singledispatchmethod
from typing import Optional, cast, overload

from gravitino.api.audit import Audit
from gravitino.api.rel.column import Column
from gravitino.api.rel.expressions.distributions.distribution import Distribution
from gravitino.api.rel.expressions.distributions.distributions import Distributions
from gravitino.api.rel.expressions.expression import Expression
from gravitino.api.rel.expressions.function_expression import FunctionExpression
from gravitino.api.rel.expressions.literals.literal import Literal
from gravitino.api.rel.expressions.literals.literals import Literals
from gravitino.api.rel.expressions.named_reference import FieldReference, NamedReference
from gravitino.api.rel.expressions.sorts.sort_order import SortOrder
from gravitino.api.rel.expressions.sorts.sort_orders import SortOrders
from gravitino.api.rel.expressions.transforms.transform import (
    SingleFieldTransform,
    Transform,
)
from gravitino.api.rel.expressions.transforms.transforms import Transforms
from gravitino.api.rel.expressions.unparsed_expression import UnparsedExpression
from gravitino.api.rel.indexes.index import Index
from gravitino.api.rel.indexes.indexes import Indexes
from gravitino.api.rel.partitions.identity_partition import IdentityPartition
from gravitino.api.rel.partitions.list_partition import ListPartition
from gravitino.api.rel.partitions.partition import Partition
from gravitino.api.rel.partitions.range_partition import RangePartition
from gravitino.api.rel.table import Table
from gravitino.api.rel.types.types import Types
from gravitino.dto.rel.column_dto import ColumnDTO
from gravitino.dto.rel.distribution_dto import DistributionDTO
from gravitino.dto.rel.expressions.field_reference_dto import FieldReferenceDTO
from gravitino.dto.rel.expressions.func_expression_dto import FuncExpressionDTO
from gravitino.dto.rel.expressions.function_arg import FunctionArg
from gravitino.dto.rel.expressions.literal_dto import LiteralDTO
from gravitino.dto.rel.expressions.unparsed_expression_dto import UnparsedExpressionDTO
from gravitino.dto.rel.indexes.index_dto import IndexDTO
from gravitino.dto.rel.partitioning.bucket_partitioning_dto import BucketPartitioningDTO
from gravitino.dto.rel.partitioning.day_partitioning_dto import DayPartitioningDTO
from gravitino.dto.rel.partitioning.function_partitioning_dto import (
    FunctionPartitioningDTO,
)
from gravitino.dto.rel.partitioning.hour_partitioning_dto import HourPartitioningDTO
from gravitino.dto.rel.partitioning.identity_partitioning_dto import (
    IdentityPartitioningDTO,
)
from gravitino.dto.rel.partitioning.list_partitioning_dto import ListPartitioningDTO
from gravitino.dto.rel.partitioning.month_partitioning_dto import MonthPartitioningDTO
from gravitino.dto.rel.partitioning.partitioning import (
    Partitioning,
    SingleFieldPartitioning,
)
from gravitino.dto.rel.partitioning.range_partitioning_dto import RangePartitioningDTO
from gravitino.dto.rel.partitioning.truncate_partitioning_dto import (
    TruncatePartitioningDTO,
)
from gravitino.dto.rel.partitioning.year_partitioning_dto import YearPartitioningDTO
from gravitino.dto.rel.partitions.identity_partition_dto import IdentityPartitionDTO
from gravitino.dto.rel.partitions.list_partition_dto import ListPartitionDTO
from gravitino.dto.rel.partitions.partition_dto import PartitionDTO
from gravitino.dto.rel.partitions.range_partition_dto import RangePartitionDTO
from gravitino.dto.rel.sort_order_dto import SortOrderDTO
from gravitino.dto.rel.table_dto import TableDTO
from gravitino.exceptions.base import IllegalArgumentException


class DTOConverters:
    """Utility class for converting between DTOs and domain objects."""

    _SINGLE_FIELD_TRANSFORM_TYPES = {
        Transforms.NAME_OF_IDENTITY: IdentityPartitioningDTO,
        Transforms.NAME_OF_YEAR: YearPartitioningDTO,
        Transforms.NAME_OF_MONTH: MonthPartitioningDTO,
        Transforms.NAME_OF_DAY: DayPartitioningDTO,
        Transforms.NAME_OF_HOUR: HourPartitioningDTO,
    }

    @staticmethod
    def from_function_arg(arg: FunctionArg) -> Expression:
        """Converts a FunctionArg DTO to an Expression.

        Args:
            arg (FunctionArg): The function argument DTO to be converted.

        Returns:
            Expression: The expression.
        """
        arg_type = arg.arg_type()
        if arg_type is FunctionArg.ArgType.LITERAL:
            dto = cast(LiteralDTO, arg)
            if dto.value() is None or dto.data_type() == Types.NullType.get():
                return Literals.NULL
            return Literals.of(dto.value(), dto.data_type())
        if arg_type is FunctionArg.ArgType.FIELD:
            dto = cast(FieldReferenceDTO, arg)
            return NamedReference.field(dto.field_name())
        if arg_type is FunctionArg.ArgType.FUNCTION:
            dto = cast(FuncExpressionDTO, arg)
            return FunctionExpression.of(
                dto.function_name(), *DTOConverters.from_function_args(dto.args())
            )
        if arg_type is FunctionArg.ArgType.UNPARSED:
            dto = cast(UnparsedExpressionDTO, arg)
            return UnparsedExpression.of(dto.unparsed_expression())
        raise IllegalArgumentException(f"Unsupported expression type: {arg}")

    @staticmethod
    def from_function_args(args: list[FunctionArg]) -> list[Expression]:
        """Converts a FunctionArg DTO to an Expression.

        Args:
            args (list[FunctionArg]): The function argument DTOs to be converted.

        Returns:
            list[Expression]: The list of expressions.
        """
        if not args:
            return Expression.EMPTY_EXPRESSION
        return [DTOConverters.from_function_arg(arg) for arg in args]

    @singledispatchmethod
    @staticmethod
    def from_dto(dto) -> object:
        raise IllegalArgumentException(f"Unsupported DTO type: {type(dto)}")

    @from_dto.register
    @staticmethod
    def _(dto: DistributionDTO) -> Distribution:
        """Converts a DistributionDTO to a Distribution.

        Args:
            dto (DistributionDTO): The distribution DTO.

        Returns:
            Distribution: The distribution.
        """
        if dto is None or DistributionDTO.NONE.equals(dto):
            return Distributions.NONE

        return Distributions.of(
            dto.strategy(),
            dto.number(),
            *DTOConverters.from_function_args(dto.args()),
        )

    @from_dto.register
    @staticmethod
    def _(dto: IndexDTO) -> Index:
        """Converts an IndexDTO to an Index.

        Args:
            dto (IndexDTO): The Index DTO to be converted.

        Returns:
            Index: The index.
        """
        return Indexes.of(dto.type(), dto.name(), dto.field_names())

    @from_dto.register
    @staticmethod
    def _(dto: SortOrderDTO) -> SortOrder:
        """Converts a SortOrderDTO to a SortOrder.

        Args:
            dto (SortOrderDTO): The sort order DTO to be converted.

        Returns:
            SortOrder: The sort order.
        """
        return SortOrders.of(
            DTOConverters.from_function_arg(dto.sort_term()),
            dto.direction(),
            dto.null_ordering(),
        )

    @from_dto.register
    @staticmethod
    def _(dto: ColumnDTO) -> Column:
        """Converts a ColumnDTO to a Column.

        Args:
            dto (ColumnDTO): The column DTO to be converted.

        Returns:
            Column: The column.
        """
        if dto.default_value() == Column.DEFAULT_VALUE_NOT_SET:
            return dto
        return Column.of(
            dto.name(),
            dto.data_type(),
            dto.comment(),
            dto.nullable(),
            dto.auto_increment(),
            DTOConverters.from_function_arg(dto.default_value()),
        )

    @from_dto.register
    @staticmethod
    def _(dto: Partitioning) -> Transform:  # pylint: disable=too-many-return-statements
        """Converts a partitioning DTO to a Transform.

        Args:
            dto (Partitioning): The partitioning DTO to be converted.

        Returns:
            Transform: The transform.
        """
        strategy = dto.strategy()
        if strategy is Partitioning.Strategy.IDENTITY:
            return Transforms.identity(cast(SingleFieldPartitioning, dto).field_name())
        if strategy is Partitioning.Strategy.YEAR:
            return Transforms.year(cast(SingleFieldPartitioning, dto).field_name())
        if strategy is Partitioning.Strategy.MONTH:
            return Transforms.month(cast(SingleFieldPartitioning, dto).field_name())
        if strategy is Partitioning.Strategy.DAY:
            return Transforms.day(cast(SingleFieldPartitioning, dto).field_name())
        if strategy is Partitioning.Strategy.HOUR:
            return Transforms.hour(cast(SingleFieldPartitioning, dto).field_name())
        if strategy is Partitioning.Strategy.BUCKET:
            bucket_partitioning_dto = cast(BucketPartitioningDTO, dto)
            return Transforms.bucket(
                bucket_partitioning_dto.num_buckets(),
                *bucket_partitioning_dto.field_names(),
            )
        if strategy is Partitioning.Strategy.TRUNCATE:
            truncate_partitioning_dto = cast(TruncatePartitioningDTO, dto)
            return Transforms.truncate(
                truncate_partitioning_dto.width(),
                truncate_partitioning_dto.field_name(),
            )
        if strategy is Partitioning.Strategy.LIST:
            list_partitioning_dto = cast(ListPartitioningDTO, dto)
            return Transforms.list(
                field_names=list_partitioning_dto.field_names(),
                assignments=[
                    cast(ListPartition, DTOConverters.from_dto(p))
                    for p in list_partitioning_dto.assignments()
                ],
            )
        if strategy is Partitioning.Strategy.RANGE:
            range_partitioning_dto = cast(RangePartitioningDTO, dto)
            return Transforms.range(
                range_partitioning_dto.field_name(),
                [
                    cast(RangePartition, DTOConverters.from_dto(p))
                    for p in range_partitioning_dto.assignments()
                ],
            )
        if strategy is Partitioning.Strategy.FUNCTION:
            function_partitioning_dto = cast(FunctionPartitioningDTO, dto)
            return Transforms.apply(
                function_partitioning_dto.function_name(),
                DTOConverters.from_function_args(function_partitioning_dto.args()),
            )
        raise IllegalArgumentException(f"Unsupported partitioning: {strategy}")

    @from_dto.register
    @staticmethod
    def _(dto: TableDTO) -> Table:
        """Converts a TableDTO to a Table.

        Args:
            dto (TableDTO): The table DTO to be converted.

        Returns:
            Table: The table.
        """

        class TableImpl(Table):  # pylint: disable=too-many-instance-attributes
            """A table implementation."""

            def __init__(
                self,
                name: str,
                columns: list[Column],
                partitioning: list[Transform],
                sort_order: list[SortOrder],
                distribution: Distribution,
                index: list[Index],
                comment: Optional[str],
                properties: dict[str, str],
                audit_info: Audit,
            ):
                self._name = name
                self._columns = columns
                self._partitioning = partitioning
                self._sort_order = sort_order
                self._distribution = distribution
                self._index = index
                self._comment = comment
                self._properties = properties
                self._audit_info = audit_info

            def name(self) -> str:
                return self._name

            def columns(self) -> list[Column]:
                return self._columns

            def partitioning(self) -> list[Transform]:
                return self._partitioning

            def sort_order(self) -> list[SortOrder]:
                return self._sort_order

            def distribution(self) -> Distribution:
                return self._distribution

            def index(self) -> list[Index]:
                return self._index

            def comment(self) -> Optional[str]:
                return self._comment

            def properties(self) -> dict[str, str]:
                return self._properties

            def audit_info(self) -> Audit:
                return self._audit_info

        return TableImpl(
            name=dto.name(),
            columns=DTOConverters.from_dtos(dto.columns()),
            partitioning=DTOConverters.from_dtos(dto.partitioning()),
            sort_order=DTOConverters.from_dtos(dto.sort_order()),
            distribution=DTOConverters.from_dto(dto.distribution()),
            index=DTOConverters.from_dtos(dto.index()),
            comment=dto.comment(),
            properties=dto.properties(),
            audit_info=dto.audit_info(),
        )

    @overload
    @staticmethod
    def from_dtos(dtos: list[ColumnDTO]) -> list[Column]: ...  # pragma: no cover

    @overload
    @staticmethod
    def from_dtos(dtos: list[IndexDTO]) -> list[Index]: ...  # pragma: no cover

    @overload
    @staticmethod
    def from_dtos(dtos: list[SortOrderDTO]) -> list[SortOrder]: ...  # pragma: no cover

    @overload
    @staticmethod
    def from_dtos(dtos: list[Partitioning]) -> list[Transform]: ...  # pragma: no cover

    @staticmethod
    def from_dtos(dtos):
        """Converts list of `ColumnDTO`, `IndexDTO`, `SortOrderDTO`, or `Partitioning`
        to the corresponding list of `Column`s, `Index`es, `SortOrder`s, or `Transform`s.

        Args:
            dtos (list[ColumnDTO] | list[IndexDTO] | list[SortOrderDTO] | list[Partitioning]):
                The DTOs to be converted.

        Returns:
            list[Column] | list[Index] | list[SortOrder] | list[Transform]:
                The list of Columns, Indexes, SortOrders, or Transforms depends on the input DTOs.
        """
        if not dtos:
            return []
        return [DTOConverters.from_dto(dto) for dto in dtos]

    @staticmethod
    def to_function_args(
        expressions: list[Expression],
    ) -> list[FunctionArg]:
        """Converts a list of Expressions to a list of FunctionArg DTOs.

        Args:
            expressions (list[Expression]): The expressions to be converted.

        Returns:
            list[FunctionArg]: The list of expression DTOs.
        """
        if not expressions:
            return []
        return [DTOConverters.to_function_arg(expr) for expr in expressions]

    @staticmethod
    def to_function_arg(expression: Expression) -> FunctionArg:
        """Converts an Expression to an FunctionArg DTO.

        Args:
            expression (Expression): The expression to be converted.

        Returns:
            FunctionArg: The expression DTO.
        """

        if isinstance(expression, FunctionArg):
            return cast(FunctionArg, expression)
        if isinstance(expression, Literal):
            if expression is Literals.NULL:
                return LiteralDTO.NULL
            return (
                LiteralDTO.builder()
                .with_value(str(expression.value()))
                .with_data_type(expression.data_type())
                .build()
            )
        if isinstance(expression, FieldReference):
            return (
                FieldReferenceDTO.builder()
                .with_field_name(expression.field_name())
                .build()
            )
        if isinstance(expression, FunctionExpression):
            return (
                FuncExpressionDTO.builder()
                .with_function_name(expression.function_name())
                .with_function_args(
                    [
                        DTOConverters.to_function_arg(arg)
                        for arg in expression.arguments()
                    ]
                )
                .build()
            )
        if isinstance(expression, UnparsedExpression):
            return (
                UnparsedExpressionDTO.builder()
                .with_unparsed_expression(expression.unparsed_expression())
                .build()
            )
        raise IllegalArgumentException(
            f"Unsupported expression type: {expression.__class__.__name__}"
        )

    @singledispatchmethod
    @staticmethod
    def to_dto(obj) -> object:
        raise IllegalArgumentException(f"Unsupported type: {type(obj)}")

    @to_dto.register
    @staticmethod
    def _(obj: Partition) -> PartitionDTO:
        if isinstance(obj, RangePartition):
            range_partition = cast(RangePartition, obj)
            return RangePartitionDTO(
                name=range_partition.name(),
                upper=cast(
                    LiteralDTO, DTOConverters.to_function_arg(range_partition.upper())
                ),
                lower=cast(
                    LiteralDTO, DTOConverters.to_function_arg(range_partition.lower())
                ),
                properties=range_partition.properties(),
            )
        if isinstance(obj, IdentityPartition):
            identity_partition = cast(IdentityPartition, obj)
            return IdentityPartitionDTO(
                name=identity_partition.name(),
                values=[
                    cast(LiteralDTO, DTOConverters.to_function_arg(v))
                    for v in identity_partition.values()
                ],
                field_names=identity_partition.field_names(),
                properties=identity_partition.properties(),
            )
        if isinstance(obj, ListPartition):
            list_partition = cast(ListPartition, obj)
            return ListPartitionDTO(
                name=list_partition.name(),
                lists=[
                    [
                        cast(LiteralDTO, DTOConverters.to_function_arg(v))
                        for v in list_item
                    ]
                    for list_item in list_partition.lists()
                ],
                properties=list_partition.properties(),
            )
        raise IllegalArgumentException(
            f"Unsupported partition type: {obj.__class__.__name__}"
        )

    @to_dto.register
    @staticmethod
    def _(obj: Column) -> ColumnDTO:
        return (
            ColumnDTO.builder()
            .with_name(obj.name())
            .with_data_type(obj.data_type())
            .with_comment(obj.comment())
            .with_nullable(obj.nullable())
            .with_auto_increment(obj.auto_increment())
            .with_default_value(
                Column.DEFAULT_VALUE_NOT_SET
                if obj.default_value() is None
                or obj.default_value() is Column.DEFAULT_VALUE_NOT_SET
                else DTOConverters.to_function_arg(obj.default_value())
            )
            .build()
        )

    @to_dto.register
    @staticmethod
    def _(obj: SortOrderDTO) -> SortOrderDTO:
        return obj

    @to_dto.register
    @staticmethod
    def _(obj: SortOrder) -> SortOrderDTO:
        return SortOrderDTO(
            sort_term=DTOConverters.to_function_arg(obj.expression()),
            direction=obj.direction(),
            null_ordering=obj.null_ordering(),
        )

    @to_dto.register
    @staticmethod
    def _(obj: IndexDTO) -> IndexDTO:
        return obj

    @to_dto.register
    @staticmethod
    def _(obj: Index) -> IndexDTO:
        return IndexDTO(
            index_type=obj.type(),
            name=obj.name(),
            field_names=obj.field_names(),
        )

    @to_dto.register
    @staticmethod
    def _(obj: Partitioning) -> Partitioning:
        return obj

    @to_dto.register
    @staticmethod
    def _(obj: Transform) -> Partitioning:
        if isinstance(obj, SingleFieldTransform):
            transform_class = DTOConverters._SINGLE_FIELD_TRANSFORM_TYPES.get(
                obj.name()
            )
            if transform_class is None:
                raise IllegalArgumentException(f"Unsupported transform: {obj.name()}")
            return transform_class(*obj.field_name())
        if isinstance(obj, Transforms.BucketTransform):
            bucket_transform = cast(Transforms.BucketTransform, obj)
            return BucketPartitioningDTO(
                bucket_transform.num_buckets(),
                *bucket_transform.field_names(),
            )
        if isinstance(obj, Transforms.TruncateTransform):
            truncate_transform = cast(Transforms.TruncateTransform, obj)
            return TruncatePartitioningDTO(
                truncate_transform.width(),
                truncate_transform.field_name(),
            )
        if isinstance(obj, Transforms.ListTransform):
            list_transform = cast(Transforms.ListTransform, obj)
            list_assignments: list[ListPartitionDTO] = [
                cast(ListPartitionDTO, DTOConverters.to_dto(assignment))
                for assignment in list_transform.assignments()
            ]
            return ListPartitioningDTO(
                field_names=list_transform.field_names(),
                assignments=list_assignments,
            )
        if isinstance(obj, Transforms.RangeTransform):
            range_transform = cast(Transforms.RangeTransform, obj)
            range_assignments: list[RangePartitionDTO] = [
                cast(RangePartitionDTO, DTOConverters.to_dto(assignment))
                for assignment in range_transform.assignments()
            ]
            return RangePartitioningDTO(
                field_name=range_transform.field_name(),
                assignments=range_assignments,
            )
        if isinstance(obj, Transforms.ApplyTransform):
            return FunctionPartitioningDTO(
                obj.name(), *DTOConverters.to_function_args(obj.arguments())
            )
        raise IllegalArgumentException(f"Unsupported transform: {obj.name()}")

    @to_dto.register
    @staticmethod
    def _(obj: Distribution) -> DistributionDTO:
        if obj is None or obj is Distributions.NONE:
            return DistributionDTO.NONE
        if isinstance(obj, DistributionDTO):
            return obj
        return DistributionDTO(
            strategy=obj.strategy(),
            number=obj.number(),
            args=[
                DTOConverters.to_function_arg(expression)
                for expression in obj.expressions()
            ],
        )

    @overload
    @staticmethod
    def to_dtos(dtos: list[Column]) -> list[ColumnDTO]: ...  # pragma: no cover

    @overload
    @staticmethod
    def to_dtos(dtos: list[Index]) -> list[IndexDTO]: ...  # pragma: no cover

    @overload
    @staticmethod
    def to_dtos(dtos: list[IndexDTO]) -> list[IndexDTO]: ...  # pragma: no cover

    @overload
    @staticmethod
    def to_dtos(dtos: list[SortOrder]) -> list[SortOrderDTO]: ...  # pragma: no cover

    @overload
    @staticmethod
    def to_dtos(dtos: list[SortOrderDTO]) -> list[SortOrderDTO]: ...  # pragma: no cover

    @overload
    @staticmethod
    def to_dtos(dtos: list[Transform]) -> list[Partitioning]: ...  # pragma: no cover

    @overload
    @staticmethod
    def to_dtos(dtos: list[Partitioning]) -> list[Partitioning]: ...  # pragma: no cover

    @staticmethod
    def to_dtos(dtos):
        if not dtos:
            return []
        return [DTOConverters.to_dto(dto) for dto in dtos]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import List

from gravitino.api.rel.expressions.expression import Expression
from gravitino.api.rel.expressions.transforms.transforms import Transforms
from gravitino.dto.rel.column_dto import ColumnDTO
from gravitino.dto.rel.partition_utils import PartitionUtils
from gravitino.dto.rel.partitioning.partitioning import Partitioning


class BucketPartitioningDTO(Partitioning):
    """Data transfer object representing bucket partitioning."""

    def __init__(self, num_buckets: int, /, *field_names: List[str]):
        self._num_buckets = num_buckets
        self._field_names = [*field_names]

    def num_buckets(self) -> int:
        """Returns the number of buckets.

        Returns:
            int: The number of buckets.
        """
        return self._num_buckets

    def field_names(self) -> List[List[str]]:
        """Returns the field names.

        Returns:
            List[List[str]]: The field names.
        """
        return self._field_names

    def strategy(self) -> Partitioning.Strategy:
        return self.Strategy.BUCKET

    def validate(self, columns: List[ColumnDTO]) -> None:
        for field_name in self._field_names:
            PartitionUtils.validate_field_existence(columns, field_name)

    def name(self) -> str:
        return self.strategy().name.lower()

    def arguments(self) -> List[Expression]:
        """Gets the arguments of the partitioning strategy.

        Returns:
            List[Expression]: The arguments of the partitioning strategy.
        """
        return Transforms.bucket(self._num_buckets, *self._field_names).arguments()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from abc import ABC, abstractmethod

from gravitino.api.tag.tag import Tag
from gravitino.api.tag.tag_change import TagChange


class TagOperations(ABC):
    """
    Interface for supporting global tag operations. This interface will provide tag listing, getting,
    creating, and other tag operations under a metalake. This interface will be mixed with
    GravitinoMetalake or GravitinoClient to provide tag operations.
    """

    @abstractmethod
    def list_tags(self) -> list[str]:
        """List all the tag names under a metalake.

        Returns:
            list[str]: The list of tag names.

        Raises:
            NoSuchMetalakeException: If the metalake does not exist.
        """
        pass

    @abstractmethod
    def list_tags_info(self) -> list[Tag]:
        """
        List tags information under a metalake.

        Returns:
            list[Tag]: The list of tag information.

        Raises:
            NoSuchMetalakeException: If the metalake does not exist.
        """
        pass

    @abstractmethod
    def get_tag(self, tag_name: str) -> Tag:
        """
        Get a tag by its name under a metalake.

        Args:
            tag_name (str): The name of the tag.

        Returns:
            Tag: The tag information.

        Raises:
            NoSuchTagException: If the tag does not exist.
        """
        pass

    @abstractmethod
    def create_tag(
        self,
        tag_name: str,
        comment: str,
        properties: dict[str, str],
    ) -> Tag:
        """
        Create a new tag under a metalake.

        Raises:
            NoSuchMetalakeException: If the metalake does not exist.
            TagAlreadyExistsException: If the tag already exists.

        Args:
            tag_name (str): The name of the tag.
            comment (str): The comment of the tag.
            properties (dict[str, str]): The properties of the tag.

        Returns:
            Tag: The tag information.
        """
        pass

    @abstractmethod
    def alter_tag(
        self,
        tag_name: str,
        *changes: TagChange,
    ) -> Tag:
        """
        Alter a tag under a metalake.

        Args:
            tag_name (str): The name of the tag.
            changes (TagChange): The changes to apply to the tag.

        Returns:
            Tag: The altered tag.

        Raises:
            NoSuchTagException: If the tag does not exist.
            NoSuchMetalakeException: If the metalake does not exist.
        """
        pass

    @abstractmethod
    def delete_tag(self, tag_name: str) -> bool:
        """
        Delete a tag under a metalake.

        Args:
            tag_name (str): The name of the tag.

        Returns:
            bool: True if the tag was deleted, False otherwise.

        Raises:
            NoSuchTagException: If the tag does not exist.
            NoSuchMetalakeException: If the metalake does not exist.
        """
        pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Dict, List

from gravitino.api.catalog import Catalog
from gravitino.api.model.model import Model
from gravitino.api.model.model_version import ModelVersion
from gravitino.api.model_change import ModelChange
from gravitino.api.model_version_change import ModelVersionChange
from gravitino.client.base_schema_catalog import BaseSchemaCatalog
from gravitino.client.generic_model import GenericModel
from gravitino.client.generic_model_version import GenericModelVersion
from gravitino.dto.audit_dto import AuditDTO
from gravitino.dto.requests.model_register_request import ModelRegisterRequest
from gravitino.dto.requests.model_update_request import ModelUpdateRequest
from gravitino.dto.requests.model_updates_request import ModelUpdatesRequest
from gravitino.dto.requests.model_version_link_request import ModelVersionLinkRequest
from gravitino.dto.requests.model_version_update_request import (
    ModelVersionUpdateRequest,
)
from gravitino.dto.requests.model_version_updates_request import (
    ModelVersionUpdatesRequest,
)
from gravitino.dto.responses.base_response import BaseResponse
from gravitino.dto.responses.drop_response import DropResponse
from gravitino.dto.responses.entity_list_response import EntityListResponse
from gravitino.dto.responses.model_response import ModelResponse
from gravitino.dto.responses.model_version_list_response import (
    ModelVersionInfoListResponse,
    ModelVersionListResponse,
)
from gravitino.dto.responses.model_version_response import ModelVersionResponse
from gravitino.dto.responses.model_version_uri_response import ModelVersionUriResponse
from gravitino.exceptions.handlers.model_error_handler import MODEL_ERROR_HANDLER
from gravitino.name_identifier import NameIdentifier
from gravitino.namespace import Namespace
from gravitino.rest.rest_utils import encode_string
from gravitino.utils import HTTPClient


class GenericModelCatalog(BaseSchemaCatalog):
    """
    The generic model catalog is a catalog that supports model and model version operations,
    for example, model register, model version link, model and model version list, etc.
    A model catalog is under the metalake.
    """

    def __init__(
        self,
        namespace: Namespace,
        name: str = None,
        catalog_type: Catalog.Type = Catalog.Type.UNSUPPORTED,
        provider: str = None,
        comment: str = None,
        properties: Dict[str, str] = None,
        audit: AuditDTO = None,
        rest_client: HTTPClient = None,
    ):
        super().__init__(
            namespace,
            name,
            catalog_type,
            provider,
            comment,
            properties,
            audit,
            rest_client,
        )

    def as_model_catalog(self):
        return self

    def list_models(self, namespace: Namespace) -> List[NameIdentifier]:
        """List the models in a schema namespace from the catalog.

        Args:
            namespace: The namespace of the schema.

        Raises:
            NoSuchSchemaException: If the schema does not exist.

        Returns:
            A list of NameIdentifier of models under the given namespace.
        """
        self._check_model_namespace(namespace)

        model_full_ns = self._model_full_namespace(namespace)
        resp = self.rest_client.get(
            self._format_model_request_path(model_full_ns),
            error_handler=MODEL_ERROR_HANDLER,
        )
        entity_list_resp = EntityListResponse.from_json(resp.body, infer_missing=True)
        entity_list_resp.validate()

        return [
            NameIdentifier.of(ident.namespace().level(2), ident.name())
            for ident in entity_list_resp.identifiers()
        ]

    def get_model(self, ident: NameIdentifier) -> Model:
        """Get a model by its identifier.

        Args:
            ident: The identifier of the model.

        Raises:
            NoSuchModelException: If the model does not exist.

        Returns:
            The model object.
        """
        self._check_model_ident(ident)

        model_full_ns = self._model_full_namespace(ident.namespace())
        resp = self.rest_client.get(
            f"{self._format_model_request_path(model_full_ns)}/{encode_string(ident.name())}",
            error_handler=MODEL_ERROR_HANDLER,
        )
        model_resp = ModelResponse.from_json(resp.body, infer_missing=True)
        model_resp.validate()

        return GenericModel(model_resp.model())

    def register_model(
        self, ident: NameIdentifier, comment: str, properties: Dict[str, str]
    ) -> Model:
        """Register a model in the catalog if the model is not existed, otherwise the
        ModelAlreadyExistsException will be thrown. The Model object will be created when the
        model is registered, users can call ModelCatalog#link_model_version to link the model
        version to the registered Model.

        Args:
            ident: The identifier of the model.
            comment: The comment of the model.
            properties: The properties of the model.

        Raises:
            ModelAlreadyExistsException: If the model already exists.
            NoSuchSchemaException: If the schema does not exist.

        Returns:
            The registered model object.
        """
        self._check_model_ident(ident)

        model_full_ns = self._model_full_namespace(ident.namespace())
        model_req = ModelRegisterRequest(
            name=ident.name(), comment=comment, properties=properties
        )
        model_req.validate()

        resp = self.rest_client.post(
            self._format_model_request_path(model_full_ns),
            model_req,
            error_handler=MODEL_ERROR_HANDLER,
        )
        model_resp = ModelResponse.from_json(resp.body, infer_missing=True)
        model_resp.validate()

        return GenericModel(model_resp.model())

    def delete_model(self, model_ident: NameIdentifier) -> bool:
        """Delete the model from the catalog. If the model does not exist, return false.
        If the model is successfully deleted, return true. The deletion of the model will also
        delete all the model versions linked to this model.

        Args:
            model_ident: The identifier of the model.

        Returns:
            True if the model is deleted successfully, False is the model does not exist.
        """
        self._check_model_ident(model_ident)

        model_full_ns = self._model_full_namespace(model_ident.namespace())
        resp = self.rest_client.delete(
            f"{self._format_model_request_path(model_full_ns)}/{encode_string(model_ident.name())}",
            error_handler=MODEL_ERROR_HANDLER,
        )
        drop_resp = DropResponse.from_json(resp.body, infer_missing=True)
        drop_resp.validate()

        return drop_resp.dropped()

    def list_model_versions(self, model_ident: NameIdentifier) -> List[int]:
        """List all the versions of the register model by NameIdentifier in the catalog.

        Args:
            model_ident: The identifier of the model.

        Raises:
            NoSuchModelException: If the model does not exist.

        Returns:
            A list of model versions.
        """
        self._check_model_ident(model_ident)

        model_full_ident = self._model_full_identifier(model_ident)
        resp = self.rest_client.get(
            f"{self._format_model_version_request_path(model_full_ident)}/versions",
            error_handler=MODEL_ERROR_HANDLER,
        )
        model_version_list_resp = ModelVersionListResponse.from_json(
            resp.body, infer_missing=True
        )
        model_version_list_resp.validate()

        return model_version_list_resp.versions()

    def list_model_version_infos(
        self, model_ident: NameIdentifier
    ) -> List[ModelVersion]:
        """List all the versions with their information of the register model by NameIdentifier in the catalog.

        Args:
            model_ident: The identifier of the model.

        Raises:
            NoSuchModelException: If the model does not exist.

        Returns:
            A list of model versions with their information.
        """
        self._check_model_ident(model_ident)
        model_full_ident = self._model_full_identifier(model_ident)

        params = {"details": "true"}
        resp = self.rest_client.get(
            f"{self._format_model_version_request_path(model_full_ident)}/versions",
            params=params,
            error_handler=MODEL_ERROR_HANDLER,
        )
        model_version_info_list_resp = ModelVersionInfoListResponse.from_json(
            resp.body, infer_missing=True
        )
        model_version_info_list_resp.validate()

        return [
            GenericModelVersion(version)
            for version in model_version_info_list_resp.versions()
        ]

    def get_model_version(
        self, model_ident: NameIdentifier, version: int
    ) -> ModelVersion:
        """Get a model version by its identifier and version.

        Args:
            model_ident: The identifier of the model.
            version: The version of the model.

        Raises:
            NoSuchModelVersionException: If the model version does not exist.

        Returns:
            The model version object.
        """
        self._check_model_ident(model_ident)

        model_full_ident = self._model_full_identifier(model_ident)
        resp = self.rest_client.get(
            f"{self._format_model_version_request_path(model_full_ident)}/versions/{version}",
            error_handler=MODEL_ERROR_HANDLER,
        )
        model_version_resp = ModelVersionResponse.from_json(
            resp.body, infer_missing=True
        )
        model_version_resp.validate()

        return GenericModelVersion(model_version_resp.model_version())

    def get_model_version_by_alias(
        self, model_ident: NameIdentifier, alias: str
    ) -> ModelVersion:
        """
        Get a model version by its identifier and alias.

        Args:
            model_ident: The identifier of the model.
            alias: The alias of the model version.

        Raises:
            NoSuchModelVersionException: If the model version does not exist.

        Returns:
            The model version object.
        """
        self._check_model_ident(model_ident)

        model_full_ident = self._model_full_identifier(model_ident)
        resp = self.rest_client.get(
            f"{self._format_model_version_request_path(model_full_ident)}/aliases/"
            f"{encode_string(alias)}",
            error_handler=MODEL_ERROR_HANDLER,
        )
        model_version_resp = ModelVersionResponse.from_json(
            resp.body, infer_missing=True
        )
        model_version_resp.validate()

        return GenericModelVersion(model_version_resp.model_version())

    def alter_model(self, model_ident: NameIdentifier, *changes: ModelChange) -> Model:
        """Alter the schema by applying the changes.
        Args:
            model_ident: The identifier of the model.
            changes: The changes to apply to the model.
        Raises:
            NoSuchSchemaException: If the schema does not exist.
            IllegalArgumentException: If the changes are invalid.
        Returns:
            The updated schema object.
        """
        self._check_model_ident(model_ident)
        model_full_ns = self._model_full_namespace(model_ident.namespace())

        update_requests = [
            GenericModelCatalog.to_model_update_request(change) for change in changes
        ]

        req = ModelUpdatesRequest(update_requests)
        req.validate()

        resp = self.rest_client.put(
            f"{self._format_model_request_path(model_full_ns)}/{encode_string(model_ident.name())}",
            req,
            error_handler=MODEL_ERROR_HANDLER,
        )
        model_response = ModelResponse.from_json(resp.body, infer_missing=True)
        model_response.validate()
        return GenericModel(model_response.model())

    def alter_model_version(
        self, model_ident: NameIdentifier, version: int, *changes: ModelVersionChange
    ):
        """
        Alter the model version by applying the changes.
        Args:
            model_ident: The identifier of the model.
            version: The version of the model version.
            changes: The changes to apply to the model version.
        Raises:
            NoSuchModelVersionException: If the model version does not exist.
            IllegalArgumentException: If the changes are invalid.
        Returns:
            The updated model version object.
        """
        self._check_model_ident(model_ident)

        model_full_ident = self._model_full_identifier(model_ident)

        update_requests = [
            GenericModelCatalog.to_model_version_update_request(change)
            for change in changes
        ]

        req = ModelVersionUpdatesRequest(update_requests)
        req.validate()

        resp = self.rest_client.put(
            f"{self._format_model_version_request_path(model_full_ident)}/versions/{version}",
            req,
            error_handler=MODEL_ERROR_HANDLER,
        )

        model_version_response = ModelVersionResponse.from_json(
            resp.body, infer_missing=True
        )
        model_version_response.validate()
        return GenericModelVersion(model_version_response.model_version())

    def alter_model_version_by_alias(
        self, model_ident: NameIdentifier, alias: str, *changes: ModelVersionChange
    ):
        """
        Alter the model version by applying the changes.
        Args:
            model_ident: The identifier of the model.
            alias: The alias of the model version.
            changes: The changes to apply to the model version.
        Raises:
            NoSuchModelVersionException: If the model version does not exist.
            IllegalArgumentException: If the changes are invalid.
        Returns:
            The updated model version object.
        """
        self._check_model_ident(model_ident)

        model_full_ident = self._model_full_identifier(model_ident)

        update_requests = [
            GenericModelCatalog.to_model_version_update_request(change)
            for change in changes
        ]

        req = ModelVersionUpdatesRequest(update_requests)
        req.validate()

        resp = self.rest_client.put(
            f"{self._format_model_version_request_path(model_full_ident)}/aliases/"
            f"{encode_string(alias)}",
            req,
            error_handler=MODEL_ERROR_HANDLER,
        )

        model_version_response = ModelVersionResponse.from_json(
            resp.body, infer_missing=True
        )
        model_version_response.validate()
        return GenericModelVersion(model_version_response.model_version())

    def link_model_version(
        self,
        model_ident: NameIdentifier,
        uri: str,
        aliases: List[str],
        comment: str,
        properties: Dict[str, str],
    ) -> None:
        """Link a new model version to the registered model object. The new model version will be
        added to the model object. If the model object does not exist, it will throw an
        exception. If the version alias already exists in the model, it will throw an exception.

        Args:
            model_ident: The identifier of the model.
            uri: The URI of the model version.
            aliases: The aliases of the model version. The aliases of the model version. The
            aliases should be unique in this model, otherwise the
            ModelVersionAliasesAlreadyExistException will be thrown. The aliases are optional and
            can be empty.
            comment: The comment of the model version.
            properties: The properties of the model version.

        Raises:
            NoSuchModelException: If the model does not exist.
            ModelVersionAliasesAlreadyExistException: If the aliases of the model version already exist.
        """
        uris = {ModelVersion.URI_NAME_UNKNOWN: uri} if uri else {}

        return self.link_model_version_with_multiple_uris(
            model_ident, uris, aliases, comment, properties
        )

    def link_model_version_with_multiple_uris(
        self,
        model_ident: NameIdentifier,
        uris: Dict[str, str],
        aliases: List[str],
        comment: str,
        properties: Dict[str, str],
    ) -> None:
        """Link a new model version to the registered model object. The new model version will be
        added to the model object. If the model object does not exist, it will throw an
        exception. If the version alias already exists in the model, it will throw an exception.

        Args:
            model_ident: The identifier of the model.
            uris: The URIs and their names of the model version.
            aliases: The aliases of the model version. The aliases of the model version. The
            aliases should be unique in this model, otherwise the
            ModelVersionAliasesAlreadyExistException will be thrown. The aliases are optional and
            can be empty.
            comment: The comment of the model version.
            properties: The properties of the model version.

        Raises:
            NoSuchModelException: If the model does not exist.
            ModelVersionAliasesAlreadyExistException: If the aliases of the model version already exist.
        """
        self._check_model_ident(model_ident)

        model_full_ident = self._model_full_identifier(model_ident)

        request = ModelVersionLinkRequest(uris, comment, aliases, properties)
        request.validate()

        resp = self.rest_client.post(
            f"{self._format_model_version_request_path(model_full_ident)}/versions",
            request,
            error_handler=MODEL_ERROR_HANDLER,
        )
        base_resp = BaseResponse.from_json(resp.body, infer_missing=True)
        base_resp.validate()

    def get_model_version_uri(
        self, model_ident: NameIdentifier, version: int, uri_name: str = None
    ):
        """Get the URI of the model artifact with a specified version number.

        Args:
            model_ident: The identifier of the model.
            version: The version of the model.
            uri_name: The name of the URI. If None, the default URI will be used.

        Raises:
            NoSuchModelVersionException: If the model version does not exist.
            NoSuchModelVersionURINameException: If the uri name does not exist.

        Returns:
            The URI of the model version.
        """
        self._check_model_ident(model_ident)

        model_full_ident = self._model_full_identifier(model_ident)
        params = {}
        if uri_name is not None:
            params["uriName"] = uri_name

        resp = self.rest_client.get(
            f"{self._format_model_version_request_path(model_full_ident)}/versions/{version}/uri",
            params=params,
            error_handler=MODEL_ERROR_HANDLER,
        )

        model_version_uri_resp = ModelVersionUriResponse.from_json(
            resp.body, infer_missing=True
        )
        model_version_uri_resp.validate()

        return model_version_uri_resp.uri()

    def get_model_version_uri_by_alias(
        self, model_ident: NameIdentifier, alias: str, uri_name: str = None
    ):
        """Get the URI of the model artifact with a specified version alias.

        Args:
            model_ident: The identifier of the model.
            alias: The alias of the model version.
            uri_name: The name of the URI. If None, the default URI will be used.

        Raises:
            NoSuchModelVersionException: If the model version does not exist.
            NoSuchModelVersionURINameException: If the uri name does not exist.

        Returns:
            The URI of the model version.
        """
        self._check_model_ident(model_ident)

        model_full_ident = self._model_full_identifier(model_ident)
        params = {}
        if uri_name is not None:
            params["uriName"] = uri_name

        resp = self.rest_client.get(
            f"{self._format_model_version_request_path(model_full_ident)}/aliases/{encode_string(alias)}/uri",
            params=params,
            error_handler=MODEL_ERROR_HANDLER,
        )

        model_version_uri_resp = ModelVersionUriResponse.from_json(
            resp.body, infer_missing=True
        )
        model_version_uri_resp.validate()

        return model_version_uri_resp.uri()

    def delete_model_version(self, model_ident: NameIdentifier, version: int) -> bool:
        """Delete the model version from the catalog. If the model version does not exist, return false.
        If the model version is successfully deleted, return true.

        Args:
            model_ident: The identifier of the model.
            version: The version of the model.

        Returns:
            True if the model version is deleted successfully, False is the model version does not exist.
        """
        self._check_model_ident(model_ident)

        model_full_ident = self._model_full_identifier(model_ident)
        resp = self.rest_client.delete(
            f"{self._format_model_version_request_path(model_full_ident)}/versions/{version}",
            error_handler=MODEL_ERROR_HANDLER,
        )
        drop_resp = DropResponse.from_json(resp.body, infer_missing=True)
        drop_resp.validate()

        return drop_resp.dropped()

    def delete_model_version_by_alias(
        self, model_ident: NameIdentifier, alias: str
    ) -> bool:
        """Delete the model version by alias from the catalog. If the model version does not exist,
        return false. If the model version is successfully deleted, return true.

        Args:
            model_ident: The identifier of the model.
            alias: The alias of the model version.

        Returns:
            True if the model version is deleted successfully, False is the model version does not exist.
        """
        self._check_model_ident(model_ident)

        model_full_ident = self._model_full_identifier(model_ident)
        resp = self.rest_client.delete(
            f"{self._format_model_version_request_path(model_full_ident)}/aliases/"
            f"{encode_string(alias)}",
            error_handler=MODEL_ERROR_HANDLER,
        )
        drop_resp = DropResponse.from_json(resp.body, infer_missing=True)
        drop_resp.validate()

        return drop_resp.dropped()

    def register_model_version(
        self,
        ident: NameIdentifier,
        uri: str,
        aliases: List[str],
        comment: str,
        properties: Dict[str, str],
    ) -> Model:
        """Register a model in the catalog if the model is not existed, otherwise the
        ModelAlreadyExistsException will be thrown. The Model object will be created when the
        model is registered, in the meantime, the model version (version 0) will also be created and
        linked to the registered model. Register a model in the catalog and link a new model
        version to the registered model.

        Args:
            ident: The identifier of the model.
            uri: The URI of the model version.
            aliases: The aliases of the model version.
            comment: The comment of the model.
            properties: The properties of the model.

        Raises:
            ModelAlreadyExistsException: If the model already exists.
            ModelVersionAliasesAlreadyExistException: If the aliases of the model version already exist.

        Returns:
            The registered model object.
        """
        uris = {ModelVersion.URI_NAME_UNKNOWN: uri} if uri else {}
        return self.register_model_version_with_multiple_uris(
            ident, uris, aliases, comment, properties
        )

    def register_model_version_with_multiple_uris(
        self,
        ident: NameIdentifier,
        uris: Dict[str, str],
        aliases: List[str],
        comment: str,
        properties: Dict[str, str],
    ) -> Model:
        """Register a model in the catalog if the model is not existed, otherwise the
        ModelAlreadyExistsException will be thrown. The Model object will be created when the
        model is registered, in the meantime, the model version (version 0) will also be created and
        linked to the registered model. Register a model in the catalog and link a new model
        version to the registered model.

        Args:
            ident: The identifier of the model.
            uris: The URIs and their names of the model version.
            aliases: The aliases of the model version.
            comment: The comment of the model.
            properties: The properties of the model.

        Raises:
            ModelAlreadyExistsException: If the model already exists.
            ModelVersionAliasesAlreadyExistException: If the aliases of the model version already exist.

        Returns:
            The registered model object.
        """
        model = self.register_model(ident, comment, properties)
        self.link_model_version_with_multiple_uris(
            ident, uris, aliases, comment, properties
        )
        return model

    @staticmethod
    def to_model_update_request(change: ModelChange):
        if isinstance(change, ModelChange.RenameModel):
            return ModelUpdateRequest.UpdateModelNameRequest(change.new_name())

        if isinstance(change, ModelChange.SetProperty):
            return ModelUpdateRequest.ModelSetPropertyRequest(
                change.property(), change.value()
            )

        if isinstance(change, ModelChange.RemoveProperty):
            return ModelUpdateRequest.ModelRemovePropertyRequest(change.property())

        if isinstance(change, ModelChange.UpdateComment):
            return ModelUpdateRequest.UpdateModelCommentRequest(change.new_comment())

        raise ValueError(f"Unknown change type: {type(change).__name__}")

    # pylint: disable=too-many-return-statements
    @staticmethod
    def to_model_version_update_request(change: ModelVersionChange):
        if isinstance(change, ModelVersionChange.UpdateComment):
            return ModelVersionUpdateRequest.UpdateModelVersionComment(
                change.new_comment()
            )

        if isinstance(change, ModelVersionChange.SetProperty):
            return ModelVersionUpdateRequest.SetModelVersionPropertyRequest(
                change.property(), change.value()
            )

        if isinstance(change, ModelVersionChange.RemoveProperty):
            return ModelVersionUpdateRequest.RemoveModelVersionPropertyRequest(
                change.property()
            )

        if isinstance(change, ModelVersionChange.UpdateUri):
            return ModelVersionUpdateRequest.UpdateModelVersionUriRequest(
                change.new_uri(), change.uri_name()
            )

        if isinstance(change, ModelVersionChange.AddUri):
            return ModelVersionUpdateRequest.AddModelVersionUriRequest(
                change.uri_name(), change.uri()
            )

        if isinstance(change, ModelVersionChange.RemoveUri):
            return ModelVersionUpdateRequest.RemoveModelVersionUriRequest(
                change.uri_name()
            )

        if isinstance(change, ModelVersionChange.UpdateAliases):
            return ModelVersionUpdateRequest.ModelVersionAliasesRequest(
                change.aliases_to_add(), change.aliases_to_remove()
            )

        raise ValueError(f"Unknown change type: {type(change).__name__}")

    def _check_model_namespace(self, namespace: Namespace):
        """Check the validity of the model namespace.

        Args:
            namespace: The namespace of the schema.

        Raises:
            IllegalNamespaceException: If the namespace is illegal.
        """
        Namespace.check(
            namespace is not None and namespace.length() == 1,
            f"Model namespace must be non-null and have 1 level, the input namespace is {namespace}",
        )

    def _check_model_ident(self, ident: NameIdentifier):
        """Check the validity of the model identifier.

        Args:
            ident: The identifier of the model.

        Raises:
            IllegalNameIdentifierException: If the identifier is illegal.
            IllegalNamespaceException: If the namespace is illegal.
        """
        NameIdentifier.check(
            ident is not None and ident.has_namespace(),
            f"Model identifier must be non-null and have a namespace, the input identifier is {ident}",
        )
        NameIdentifier.check(
            ident.name() is not None and len(ident.name()) > 0,
            f"Model name must be non-null and non-empty, the input name is {ident.name()}",
        )
        self._check_model_namespace(ident.namespace())

    def _format_model_request_path(self, model_ns: Namespace) -> str:
        """Format the model request path.

        Args:
            model_ns: The namespace of the model.

        Returns:
            The formatted model request path.
        """
        schema_ns = Namespace.of(model_ns.level(0), model_ns.level(1))
        return (
            f"{BaseSchemaCatalog.format_schema_request_path(schema_ns)}/"
            f"{encode_string(model_ns.level(2))}/models"
        )

    def _format_model_version_request_path(self, model_ident: NameIdentifier) -> str:
        """Format the model version request path.

        Args:
            model_ident: The identifier of the model.

        Returns:
            The formatted model version request path.
        """
        return (
            f"{self._format_model_request_path(model_ident.namespace())}"
            f"/{encode_string(model_ident.name())}"
        )

    def _model_full_namespace(self, model_namespace: Namespace) -> Namespace:
        """Get the full namespace of the model.

        Args:
            model_namespace: The namespace of the model.

        Returns:
            The full namespace of the model.
        """
        return Namespace.of(
            self._catalog_namespace.level(0), self.name(), model_namespace.level(0)
        )

    def _model_full_identifier(self, model_ident: NameIdentifier) -> NameIdentifier:
        """Get the full identifier of the model.

        Args:
            model_ident: The identifier of the model.

        Returns:
            The full identifier of the model.
        """
        return NameIdentifier.builder(
            self._model_full_namespace(model_ident.namespace()), model_ident.name()
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from abc import ABC, abstractmethod
from contextlib import suppress
from typing import Optional

from gravitino.api.rel.column import Column
from gravitino.api.rel.expressions.distributions.distribution import Distribution
from gravitino.api.rel.expressions.sorts.sort_order import SortOrder
from gravitino.api.rel.expressions.transforms.transform import Transform
from gravitino.api.rel.indexes.index import Index
from gravitino.api.rel.table import Table
from gravitino.exceptions.base import (
    NoSuchTableException,
    UnsupportedOperationException,
)
from gravitino.name_identifier import NameIdentifier
from gravitino.namespace import Namespace


class TableCatalog(ABC):
    """The `TableCatalog` interface defines the public API for managing tables in a schema.

    If the catalog implementation supports tables, it must implement this interface.
    """

    @abstractmethod
    def list_tables(self, namespace: Namespace) -> list[NameIdentifier]:
        """List the tables in a namespace from the catalog.

        Args:
            namespace (Namespace): A namespace.

        Returns:
            list[NameIdentifier]: An array of table identifiers in the namespace.

        Raises:
            NoSuchSchemaException: If the schema does not exist.
        """

    @abstractmethod
    def load_table(self, identifier: NameIdentifier) -> Table:
        """Load table metadata by `NameIdentifier` from the catalog.

        Args:
            identifier (NameIdentifier): A table identifier.

        Returns:
            Table: The table metadata.

        Raises:
            NoSuchTableException: If the table does not exist.
        """

    def table_exists(self, identifier: NameIdentifier) -> bool:
        """Check if a table exists using an `NameIdentifier` from the catalog.

        Args:
            identifier (NameIdentifier): A table identifier.

        Returns:
            bool: `True` If the table exists, `False` otherwise.
        """
        with suppress(NoSuchTableException):
            self.load_table(identifier)
            return True
        return False

    @abstractmethod
    def create_table(
        self,
        identifier: NameIdentifier,
        columns: list[Column],
        comment: Optional[str] = None,
        properties: Optional[dict[str, str]] = None,
        partitioning: Optional[list[Transform]] = None,
        distribution: Optional[Distribution] = None,
        sort_orders: Optional[list[SortOrder]] = None,
        indexes: Optional[list[Index]] = None,
    ) -> Table:
        """Create a table in the catalog.

        Args:
            identifier (NameIdentifier):
                A table identifier.
            columns (list[Column]):
                The columns of the new table.
            comment (str, optional):
                The table comment. Defaults to `None`.
            properties (dict[str, str], optional):
                The table properties. Defaults to `None`.
            partitioning (Optional[list[Transform]], optional):
                The table partitioning. Defaults to None.
            distribution (Optional[Distribution], optional):
                The distribution of the table. Defaults to `None`.
            sort_orders (Optional[list[SortOrder]], optional):
                The sort orders of the table. Defaults to `None`.
            indexes (Optional[list[Index]], optional):
                The table indexes. Defaults to `None`.

        Raises:
            NoSuchSchemaException:
                If the schema does not exist.
            TableAlreadyExistsException:
                If the table already exists.

        Returns:
            Table:
                The created table metadata.
        """

    @abstractmethod
    def alter_table(self, identifier: NameIdentifier, *changes) -> Table:
        """Alter a table in the catalog.

        Args:
            identifier (NameIdentifier): A table identifier.
            *changes:
                Table changes (defined in class `TableChange`) to apply to the table.

        Returns:
            Table: The updated table metadata.

        Raises:
            NoSuchTableException:
                If the table does not exist.
            IllegalArgumentException:
                If the change is rejected by the implementation.
        """

    @abstractmethod
    def drop_table(self, identifier: NameIdentifier) -> bool:
        """Drop a table from the catalog.

        Removes both the metadata and the directory associated with the table from the
        file system if the table is not an external table. In case of an external table,
        only the associated metadata is removed.

        Args:
            identifier (NameIdentifier): A table identifier.

        Returns:
            bool: `True` if the table is dropped, `False` if the table does not exist.
        """

    def purge_table(self, identifier: NameIdentifier) -> bool:
        """Drop a table from the catalog and completely remove its data.

        Removes both the metadata and the directory associated with the table completely
        and skipping trash. If the table is an external table or the catalogs don't support
        purge table, `UnsupportedOperationException` is thrown. If the catalog supports to
        purge a table, this method should be overridden. The default implementation throws
        an `UnsupportedOperationException`.

        Args:
            identifier (NameIdentifier): A table identifier.

        Raises:
            UnsupportedOperationException: If the catalog does not support to purge a table.

        Returns:
            bool: `True` if the table is purged, `False` if the table does not exist.
        """
        raise UnsupportedOperationException("purgeTable not supported.")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from gravitino.exceptions.base import IllegalArgumentException


class Precondition:
    @staticmethod
    def check_argument(expression_result: bool, error_message: str):
        """Ensures the truth of an expression involving one or more parameters
        to the calling method.

        Args:
          expression_result: A boolean expression.
          error_message: The error message to use if the check fails.
        Raises:
          IllegalArgumentException – if expression is false
        """
        if not expression_result:
            raise IllegalArgumentException(error_message)

    @staticmethod
    def check_string_not_empty(check_string: str, error_message: str):
        """Ensures the string is not empty.

        Args:
          check_string: The string to check.
          error_message: The error message to use if the check fails.
        Raises:
          IllegalArgumentException – if the check fails.
        """
        Precondition.check_argument(
            check_string is not None and check_string.strip() != "", error_message
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from abc import ABC, abstractmethod
from collections import Counter
from collections.abc import Collection

from gravitino.api.authorization.privileges import Privilege
from gravitino.api.metadata_object import MetadataObject
from gravitino.api.metadata_objects import MetadataObjects


class SecurableObject(MetadataObject, ABC):
    """
    A securable object is an entity on which access control can be granted.
    Unless explicitly granted, access is denied.

    Apache Gravitino organizes securable objects in a tree structure.
    Each securable object contains three attributes: parent, name, and type.

    Supported types:
        - CATALOG
        - SCHEMA
        - TABLE
        - FILESET
        - TOPIC
        - METALAKE

    Use the helper class `SecurableObjects` to construct the securable object you need.

    In RESTful APIs, you can reference a securable object using its full name and type.

    Examples
    --------
    Catalog:
        - Python code:
            SecurableObjects.catalog("catalog1")
        - REST API:
            full_name="catalog1", type="CATALOG"

    Schema:
        - Python code:
            SecurableObjects.schema("catalog1", "schema1")
        - REST API:
            full_name="catalog1.schema1", type="SCHEMA"

    Table:
        - Python code:
            SecurableObjects.table("catalog1", "schema1", "table1")
        - REST API:
            full_name="catalog1.schema1.table1", type="TABLE"

    Topic:
        - Python code:
            SecurableObjects.topic("catalog1", "schema1", "topic1")
        - REST API:
            full_name="catalog1.schema1.topic1", type="TOPIC"

    Fileset:
        - Python code:
            SecurableObjects.fileset("catalog1", "schema1", "fileset1")
        - REST API:
            full_name="catalog1.schema1.fileset1", type="FILESET"

    Metalake:
        - Python code:
            SecurableObjects.metalake("metalake1")
        - REST API:
            full_name="metalake1", type="METALAKE"

    Notes
    -----
    - To represent “all catalogs”, you can use the metalake as the root object.
    - To grant a privilege on all children, you can assign it to their common parent.
      For example, to grant READ TABLE on all tables under `catalog1.schema1`,
      simply grant READ TABLE on the schema object itself.
    """

    @abstractmethod
    def privileges(self) -> list["Privilege"]:
        """The privileges of the securable object. For example: If the securable object is a table, the
        privileges could be `READ TABLE`, `WRITE TABLE`, etc. If a schema has the privilege of `LOAD
        TABLE`. It means the role can load all tables of the schema.

        returns:
            The privileges of the securable object.
        """
        raise NotImplementedError()


class SecurableObjects:
    """The helper class for SecurableObject."""

    class SecurableObjectImpl(MetadataObjects.MetadataObjectImpl, SecurableObject):
        def __init__(
            self,
            parent: str,
            name: str,
            type_: MetadataObject.Type,
            privileges: list[Privilege],
        ) -> None:
            super().__init__(name, type_, parent)
            self._privileges = list(set(privileges))

        @staticmethod
        def is_equal_collection(c1: Collection, c2: Collection) -> bool:
            if c1 is c2:
                return True
            if c1 is None or c2 is None:
                return False

            return Counter(c1) == Counter(c2)

        def __hash__(self) -> int:
            return hash((super().__hash__(), self._privileges))

        def __str__(self) -> str:
            privileges_str = ",".join(
                f"[{p.simple_string()}]" for p in self._privileges
            )

            return (
                f"SecurableObject: [fullName={self.full_name()}], "
                f"[type={self.type()}], [privileges={privileges_str}]"
            )

        def __eq__(self, other: object) -> bool:
            if other is self:
                return True
            if not isinstance(other, SecurableObject):
                return False

            return super().__eq__(
                other
            ) and SecurableObjects.SecurableObjectImpl.is_equal_collection(
                self.privileges(), other.privileges()
            )

        def privileges(self) -> list[Privilege]:
            return self._privileges

    @staticmethod
    def of_metalake(
        metalake: str,
        privileges: list[Privilege],
    ) -> SecurableObjectImpl:
        """
        Create the metalake SecurableObject with the given metalake name and privileges.

        Args:
            metalake (str): The metalake name.
            privileges (list[Privilege]): The privileges of the metalake.

        Returns:
            SecurableObjectImpl: The created metalake SecurableObject.
        """
        return SecurableObjects.of(
            MetadataObject.Type.METALAKE,
            [metalake],
            privileges,
        )

    @staticmethod
    def of_catalog(
        catalog: str,
        privileges: list[Privilege],
    ) -> SecurableObjectImpl:
        """
        Create the catalog SecurableObject with the given catalog name and privileges.

        Args:
            catalog (str): The catalog name
            privileges (list[Privilege]): The privileges of the catalog.

        Returns:
            SecurableObjectImpl: The created catalog SecurableObject.
        """
        return SecurableObjects.of(
            MetadataObject.Type.CATALOG,
            [catalog],
            privileges,
        )

    @staticmethod
    def of_schema(
        catalog: SecurableObject,
        schema: str,
        privileges: list[Privilege],
    ) -> SecurableObjectImpl:
        """
        Create the schema SecurableObject with the given securable catalog object, schema name and privileges.

        Args:
            catalog (SecurableObject): The catalog securable object.
            schema (str): The schema name.
            privileges (list[Privilege]): The privileges of the schema.

        Returns:
            SecurableObjectImpl: The created schema SecurableObject.
        """
        return SecurableObjects.of(
            MetadataObject.Type.SCHEMA,
            [catalog.full_name(), schema],
            privileges,
        )

    @staticmethod
    def of_table(
        schema: SecurableObject,
        table: str,
        privileges: list[Privilege],
    ) -> SecurableObjectImpl:
        """
        Create the table SecurableObject with the given securable schema object, table name and privileges.

        Args:
            schema (SecurableObject): The schema securable object.
            table (str): The table name.
            privileges (list[Privilege]): The privileges of the table.

        Returns:
            SecurableObjectImpl: The created table SecurableObject.
        """
        names = [*schema.full_name().split("."), table]
        return SecurableObjects.of(
            MetadataObject.Type.TABLE,
            names,
            privileges,
        )

    @staticmethod
    def of_topic(
        schema: SecurableObject,
        topic: str,
        privileges: list[Privilege],
    ) -> SecurableObjectImpl:
        """
        Create the topic SecurableObject with the given securable schema object ,topic name and privileges.

        Args:
            schema (SecurableObject): The schema securable object
            topic (str): The topic name
            privileges (list[Privilege]): The privileges of the topic

        Returns:
            SecurableObjectImpl: The created topic SecurableObject
        """
        names = [*schema.full_name().split("."), topic]
        return SecurableObjects.of(
            MetadataObject.Type.TOPIC,
            names,
            privileges,
        )

    @staticmethod
    def of_fileset(
        schema: SecurableObject,
        fileset: str,
        privileges: list[Privilege],
    ) -> SecurableObjectImpl:
        """
        Create the model SecurableObject with the given securable schema object, fileset name and

        Args:
            schema (SecurableObject): The schema securable object
            fileset (str): The fileset name
            privileges (list[Privilege]): The privileges of the fileset

        Returns:
            SecurableObjectImpl: The created fileset SecurableObject.
        """
        names = [*schema.full_name().split("."), fileset]
        return SecurableObjects.of(
            MetadataObject.Type.FILESET,
            names,
            privileges,
        )

    @staticmethod
    def of_model(
        schema: SecurableObject,
        model: str,
        privileges: list[Privilege],
    ) -> SecurableObjectImpl:
        """
        Create the model SecurableObject with the given securable schema object, model name and

        Args:
            schema (SecurableObject): The schema securable object.
            model (str): The model name.
            privileges (list[Privilege]): The privileges of the fileset.

        Returns:
            SecurableObjectImpl: The created model SecurableObject.
        """
        names = [*schema.full_name().split("."), model]
        return SecurableObjects.of(
            MetadataObject.Type.MODEL,
            names,
            privileges,
        )

    @staticmethod
    def of_tag(
        tag_name: str,
        privileges: list[Privilege],
    ) -> SecurableObjectImpl:
        """
        Create the tag SecurableObject with the given tag name and privileges.

        Args:
            tag_name (str): The tag name
            privileges (list[Privilege]): The privileges of the tag

        Returns:
            SecurableObjectImpl: The created tag SecurableObject.
        """
        return SecurableObjects.of(
            MetadataObject.Type.TAG,
            [tag_name],
            privileges,
        )

    @staticmethod
    def of_policy(
        policy: str,
        privileges: list[Privilege],
    ) -> SecurableObjectImpl:
        """
        Create the policy SecurableObject with the given policy name and privileges.

        Args:
            policy (str): The policy name
            privileges (list[Privilege]): The privileges of the policy

        Returns:
            SecurableObjectImpl: _descriThe created policy SecurableObject.
        """
        return SecurableObjects.of(
            MetadataObject.Type.POLICY,
            [policy],
            privileges,
        )

    @staticmethod
    def of_job_template(
        job_template: str,
        privileges: list[Privilege],
    ) -> SecurableObjectImpl:
        """
        Create the job template SecurableObject with the given job template name and privileges.

        Args:
            job_template (str): The job template name
            privileges (list[Privilege]): The privileges of the job template

        Returns:
            SecurableObjectImpl: The created job template SecurableObject.
        """
        return SecurableObjects.of(
            MetadataObject.Type.JOB_TEMPLATE,
            [job_template],
            privileges,
        )

    @staticmethod
    def parse(
        full_name: str,
        type_: MetadataObject.Type,
        privileges: list[Privilege],
    ) -> SecurableObject:
        """
        Create a SecurableObject from the given full name.

        Args:
            full_name (str): The full name of securable object.
            type_ (MetadataObject.Type): The securable object type.
            privileges (list[Privilege]): The securable object privileges.

        Returns:
            SecurableObject: The created SecurableObject.
        """
        metadata_object = MetadataObjects.parse(full_name, type_)
        return SecurableObjects.SecurableObjectImpl(
            metadata_object.parent(),
            metadata_object.name(),
            type_,
            privileges,
        )

    @staticmethod
    def of(
        type_: MetadataObject.Type,
        names: list[str],
        privileges: list[Privilege],
    ) -> SecurableObject:
        """
        Create the SecurableObject with the given names.

        Args:
            type_ (MetadataObject.Type): The securable object type.
            names (list[str]): The names of the securable object.
            privileges (list[Privilege]): The securable object privileges.

        Returns:
            SecurableObject: he created SecurableObject
        """
        metadata_object = MetadataObjects.of(names, type_)
        return SecurableObjects.SecurableObjectImpl(
            metadata_object.parent(),
            metadata_object.name(),
            type_,
            privileges,
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from abc import abstractmethod
from enum import Enum, unique
from typing import TYPE_CHECKING, ClassVar, List

from gravitino.api.rel.expressions.expression import Expression
from gravitino.dto.rel.partition_utils import PartitionUtils

if TYPE_CHECKING:
    from gravitino.dto.rel.column_dto import ColumnDTO


class FunctionArg(Expression):
    """An argument of a function."""

    EMPTY_ARGS: ClassVar[List[FunctionArg]] = []

    @abstractmethod
    def arg_type(self) -> ArgType:
        """Arguments type of the function.

        Returns:
            ArgType: The type of this argument.
        """
        pass

    def validate(self, columns: List[ColumnDTO]) -> None:
        """Validates the function argument.

        Args:
            columns (List[ColumnDTO]): The columns of the table.

        Raises:
            IllegalArgumentException: If the function argument is invalid.
        """
        validate_field_existence = PartitionUtils.validate_field_existence
        for ref in self.references():
            validate_field_existence(columns, ref.field_name())

    @unique
    class ArgType(str, Enum):
        """The type of the argument.

        The supported types are:

        - `LITERAL`: A literal argument.
        - `FIELD`: A field argument.
        - `FUNCTION`: A function argument.
        - `UNPARSED`: An unparsed argument.
        """

        LITERAL = "literal"
        FIELD = "field"
        FUNCTION = "function"
        UNPARSED = "unparsed"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import json
from dataclasses import dataclass, field

from dataclasses_json import config
from gravitino.exceptions.base import IllegalArgumentException

from .base_response import BaseResponse
from ..job.job_template_dto import JobTemplateDTO


@dataclass
class JobTemplateResponse(BaseResponse):
    """Represents a response containing a single job template."""

    _job_template: JobTemplateDTO = field(metadata=config(field_name="jobTemplate"))

    def validate(self):
        """Validates the response data.

        Raises:
            IllegalArgumentException: If the job template is not set or invalid.
        """
        super().validate()

        if self._job_template is None:
            raise IllegalArgumentException('"jobTemplate" must not be None')

        self._job_template.validate()

    def job_template(self) -> JobTemplateDTO:
        """Returns the job template from the response."""
        return self._job_template

    @classmethod
    def from_json(
        cls, s: str, infer_missing: bool = False, **kwargs
    ) -> "JobTemplateResponse":
        """Deserialize JSON string into a JobTemplateResponse object."""
        data = json.loads(s)
        job_template_data = JobTemplateDTO.from_json(
            json.dumps(data["jobTemplate"]), infer_missing=infer_missing, **kwargs
        )
        return cls(_job_template=job_template_data, _code=0)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from contextlib import suppress
from types import MappingProxyType
from typing import Any, Dict, Final, cast

from gravitino.api.rel.types.json_serdes.base import JsonSerializable
from gravitino.dto.rel.expressions.json_serdes._helper.serdes_utils import (
    SerdesUtils as ExpressionSerdesUtils,
)
from gravitino.dto.rel.partitioning.bucket_partitioning_dto import (
    BucketPartitioningDTO,
)
from gravitino.dto.rel.partitioning.day_partitioning_dto import DayPartitioningDTO
from gravitino.dto.rel.partitioning.function_partitioning_dto import (
    FunctionPartitioningDTO,
)
from gravitino.dto.rel.partitioning.hour_partitioning_dto import HourPartitioningDTO
from gravitino.dto.rel.partitioning.identity_partitioning_dto import (
    IdentityPartitioningDTO,
)
from gravitino.dto.rel.partitioning.list_partitioning_dto import ListPartitioningDTO
from gravitino.dto.rel.partitioning.month_partitioning_dto import MonthPartitioningDTO
from gravitino.dto.rel.partitioning.partitioning import (
    Partitioning,
    SingleFieldPartitioning,
)
from gravitino.dto.rel.partitioning.range_partitioning_dto import RangePartitioningDTO
from gravitino.dto.rel.partitioning.truncate_partitioning_dto import (
    TruncatePartitioningDTO,
)
from gravitino.dto.rel.partitioning.year_partitioning_dto import YearPartitioningDTO
from gravitino.dto.rel.partitions.json_serdes._helper.serdes_utils import (
    SerdesUtils as PartitionSerdesUtils,
)
from gravitino.dto.rel.partitions.list_partition_dto import ListPartitionDTO
from gravitino.dto.rel.partitions.range_partition_dto import RangePartitionDTO
from gravitino.utils.precondition import Precondition
from gravitino.utils.serdes import SerdesUtilsBase


class PartitioningSerdes(SerdesUtilsBase, JsonSerializable[Partitioning]):
    """Custom JSON serializer/deserializer for PartitionDTO objects."""

    _SINGLE_FIELD_PARTITIONING: Final[MappingProxyType] = MappingProxyType(
        {
            Partitioning.Strategy.IDENTITY: IdentityPartitioningDTO,
            Partitioning.Strategy.YEAR: YearPartitioningDTO,
            Partitioning.Strategy.MONTH: MonthPartitioningDTO,
            Partitioning.Strategy.DAY: DayPartitioningDTO,
            Partitioning.Strategy.HOUR: HourPartitioningDTO,
        }
    )

    @classmethod
    def serialize(cls, data_type: Partitioning) -> Dict[str, Any]:
        """Serialize the given PartitionDTO object.

        Args:
            data_type (Partitioning): The PartitionDTO objects.

        Returns:
            Dict[str, Any]: The serialized result.

        Raises:
            IOError: If partitioning strategy is unknown.
        """

        strategy = data_type.strategy()
        result = {cls.STRATEGY: strategy.name.lower()}

        if strategy in cls._SINGLE_FIELD_PARTITIONING:
            dto = cast(SingleFieldPartitioning, data_type)
            return {**result, cls.FIELD_NAME: dto.field_name()}
        if strategy is Partitioning.Strategy.BUCKET:
            dto = cast(BucketPartitioningDTO, data_type)
            return {
                **result,
                cls.NUM_BUCKETS: dto.num_buckets(),
                cls.FIELD_NAMES: dto.field_names(),
            }
        if strategy is Partitioning.Strategy.TRUNCATE:
            dto = cast(TruncatePartitioningDTO, data_type)
            return {
                **result,
                cls.WIDTH: dto.width(),
                cls.FIELD_NAME: dto.field_name(),
            }
        if strategy is Partitioning.Strategy.LIST:
            dto = cast(ListPartitioningDTO, data_type)
            return {
                **result,
                cls.FIELD_NAMES: dto.field_names(),
                cls.ASSIGNMENTS_NAME: [
                    PartitionSerdesUtils.write_partition(list_partition_dto)
                    for list_partition_dto in dto.assignments()
                ],
            }
        if strategy is Partitioning.Strategy.RANGE:
            dto = cast(RangePartitioningDTO, data_type)
            return {
                **result,
                cls.FIELD_NAME: dto.field_name(),
                cls.ASSIGNMENTS_NAME: [
                    PartitionSerdesUtils.write_partition(range_partition_dto)
                    for range_partition_dto in dto.assignments()
                ],
            }
        if strategy is Partitioning.Strategy.FUNCTION:
            dto = cast(FunctionPartitioningDTO, data_type)
            return {
                **result,
                cls.FUNCTION_NAME: dto.function_name(),
                cls.FUNCTION_ARGS: [
                    ExpressionSerdesUtils.write_function_arg(function_arg)
                    for function_arg in dto.args()
                ],
            }

        raise IOError(f"Unknown partitioning strategy: {strategy}")

    @classmethod
    def deserialize(cls, data: Dict[str, Any]) -> Partitioning:
        """Deserialize the given JSON data to the corresponding Partitioning object.

        Args:
            data (Dict[str, Any]): The JSON data object to be deserialized.

        Returns:
            Partitioning: The deserialized result.

        Raises:
            IllegalArgumentException: If there's illegal arguments in the given JSON data.
        """

        Precondition.check_argument(
            isinstance(data, dict) and len(data) > 0,
            f"Cannot parse partitioning from invalid JSON: {data}",
        )
        Precondition.check_argument(
            cls.STRATEGY in data,
            f"Cannot parse partitioning from missing strategy: {data}",
        )
        strategy = None
        with suppress(ValueError):
            strategy = Partitioning.Strategy(data[cls.STRATEGY].lower())

        if strategy in cls._SINGLE_FIELD_PARTITIONING:
            return cls._SINGLE_FIELD_PARTITIONING[strategy](
                *data.get(cls.FIELD_NAME, [])
            )
        if strategy is Partitioning.Strategy.BUCKET:
            return BucketPartitioningDTO(
                int(data[cls.NUM_BUCKETS]),
                *data.get(cls.FIELD_NAMES, []),
            )
        if strategy is Partitioning.Strategy.TRUNCATE:
            return TruncatePartitioningDTO(
                int(data[cls.WIDTH]),
                data.get(cls.FIELD_NAME, []),
            )
        if strategy is Partitioning.Strategy.LIST:
            field_names = data[cls.FIELD_NAMES]
            assignments_data = data.get(cls.ASSIGNMENTS_NAME, [])
            Precondition.check_argument(
                isinstance(assignments_data, list),
                f"Cannot parse list partitioning from non-array assignments: {assignments_data}",
            )
            assignments = []
            for assignment in assignments_data:
                partition_dto = PartitionSerdesUtils.read_partition(assignment)
                Precondition.check_argument(
                    isinstance(partition_dto, ListPartitionDTO),
                    f"Cannot parse list partitioning from non-list assignment: {assignment}",
                )
                assignments.append(partition_dto)
            return ListPartitioningDTO(field_names, assignments)
        if strategy is Partitioning.Strategy.RANGE:
            fields = data[cls.FIELD_NAME]
            assignments_data = data.get(cls.ASSIGNMENTS_NAME, [])
            Precondition.check_argument(
                isinstance(assignments_data, list),
                f"Cannot parse range partitioning from non-array assignments: {assignments_data}",
            )
            assignments = []
            for assignment in assignments_data:
                partition_dto = PartitionSerdesUtils.read_partition(assignment)
                Precondition.check_argument(
                    isinstance(partition_dto, RangePartitionDTO),
                    f"Cannot parse range partitioning from non-range assignment: {assignment}",
                )
                assignments.append(partition_dto)
            return RangePartitioningDTO(fields, assignments)
        if strategy is Partitioning.Strategy.FUNCTION:
            function_name = data[cls.FUNCTION_NAME]
            Precondition.check_argument(
                data.get(cls.FUNCTION_ARGS) is not None,
                f"Cannot parse function partitioning from missing function args: {data}",
            )
            args = [
                ExpressionSerdesUtils.read_function_arg(arg)
                for arg in data[cls.FUNCTION_ARGS]
            ]
            return FunctionPartitioningDTO(function_name, *args)

        raise IOError(f"Unknown partitioning strategy: {data[cls.STRATEGY]}")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional

from gravitino.api.rel.expressions.expression import Expression
from gravitino.api.rel.expressions.function_expression import FunctionExpression
from gravitino.api.rel.types.type import Type
from gravitino.api.tag.supports_tags import SupportsTags
from gravitino.exceptions.base import UnsupportedOperationException
from gravitino.utils.precondition import Precondition


class Column(ABC):
    """An interface representing a column of a `Table`.

    It defines basic properties of a column, such as name and data type.

    Catalog implementation needs to implement it. They should consume it in APIs like
    `TableCatalog.createTable(NameIdentifier, List[Column], str, Dict)`, and report it in
    `Table.columns()` a default value and a generation expression.
    """

    DEFAULT_VALUE_NOT_SET = Expression.EMPTY_EXPRESSION
    """A default value that indicates the default value is not set. This is used in `default_value()`"""

    DEFAULT_VALUE_OF_CURRENT_TIMESTAMP: Expression = FunctionExpression.of(
        "current_timestamp"
    )
    """
    A default value that indicates the default value will be set to the current timestamp.
    This is used in `default_value()`
    """

    @abstractmethod
    def name(self) -> str:
        """Get the name of this column.

        Returns:
            str: The name of this column.
        """
        pass

    @abstractmethod
    def data_type(self) -> Type:
        """Get the name of this column.

        Returns:
            Type: The data type of this column.
        """
        pass

    @abstractmethod
    def comment(self) -> Optional[str]:
        """Get the comment of this column.

        Returns:
            Optional[str]: The comment of this column, `None` if not specified.
        """
        pass

    @abstractmethod
    def nullable(self) -> bool:
        """Indicate if this column may produce null values.

        Returns:
            bool: `True` if this column may produce null values. Default is `True`.
        """
        return True

    @abstractmethod
    def auto_increment(self) -> bool:
        """Indicate if this column is an auto-increment column.

        Returns:
            bool: `True` if this column is an auto-increment column. Default is `False`.
        """
        return False

    @abstractmethod
    def default_value(self) -> Expression:
        """Get the default value of this column

        Returns:
            Expression:
                The default value of this column, `Column.DEFAULT_VALUE_NOT_SET` if not specified.
        """
        pass

    def supports_tags(self) -> SupportsTags:
        """Return the `SupportsTags` if the column supports tag operations.

        Raises:
            UnsupportedOperationException: if the column does not support tag operations.

        Returns:
            SupportsTags: the `SupportsTags` if the column supports tag operations.
        """
        raise UnsupportedOperationException("Column does not support tag operations.")

    @staticmethod
    def of(
        name: str,
        data_type: Type,
        comment: Optional[str] = None,
        nullable: bool = True,
        auto_increment: bool = False,
        default_value: Optional[Expression] = None,
    ) -> ColumnImpl:
        """Create a `Column` instance.

        Args:
            name (str):
                The name of the column.
            data_type (Type):
                The data type of the column.
            comment (Optional[str], optional):
                The comment of the column. Defaults to `None`.
            nullable (bool, optional):
                `True` if the column may produce null values. Defaults to `True`.
            auto_increment (bool, optional):
                `True` if the column is an auto-increment column. Defaults to `False`.
            default_value (Optional[Expression], optional):
                The default value of this column, `Column.DEFAULT_VALUE_NOT_SET` if `None`. Defaults to `None`.

        Returns:
            ColumnImpl: A `Column` instance.
        """
        return ColumnImpl(
            name=name,
            data_type=data_type,
            comment=comment,
            nullable=nullable,
            auto_increment=auto_increment,
            default_value=(
                Column.DEFAULT_VALUE_NOT_SET if default_value is None else default_value
            ),
        )


class ColumnImpl(Column):
    """The implementation of `Column` for users to use API."""

    def __init__(
        self,
        name: str,
        data_type: Type,
        comment: Optional[str],
        nullable: bool,
        auto_increment: bool,
        default_value: Optional[Expression],
    ):
        Precondition.check_string_not_empty(name, "Column name cannot be null")
        Precondition.check_argument(
            data_type is not None, "Column data type cannot be null"
        )
        self._name = name
        self._data_type = data_type
        self._comment = comment
        self._nullable = nullable
        self._auto_increment = auto_increment
        self._default_value = default_value

    def name(self) -> str:
        return self._name

    def data_type(self) -> Type:
        return self._data_type

    def comment(self) -> Optional[str]:
        return self._comment

    def nullable(self) -> bool:
        return self._nullable

    def auto_increment(self) -> bool:
        return self._auto_increment

    def default_value(self) -> Expression:
        return self._default_value

    def __eq__(self, other: ColumnImpl) -> bool:
        if not isinstance(other, ColumnImpl):
            return False
        return all(
            [
                self._name == other.name(),
                self._data_type == other.data_type(),
                self._comment == other.comment(),
                self._nullable == other.nullable(),
                self._auto_increment == other.auto_increment(),
                self._default_value == other.default_value(),
            ]
        )

    def __hash__(self):
        return hash(
            (
                self._name,
                self._data_type,
                self._comment,
                self._nullable,
                self._auto_increment,
                (
                    tuple(self._default_value)
                    if self._default_value is Column.DEFAULT_VALUE_NOT_SET
                    else self._default_value
                ),
            )
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from typing import List

from gravitino.api.rel.expressions.named_reference import NamedReference
from gravitino.dto.rel.expressions.function_arg import FunctionArg


class FieldReferenceDTO(NamedReference, FunctionArg):
    """Data transfer object representing a field reference."""

    def __init__(self, field_name: List[str]):
        self._field_name = field_name

    def field_name(self) -> List[str]:
        return self._field_name

    def arg_type(self) -> FunctionArg.ArgType:
        return self.ArgType.FIELD

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FieldReferenceDTO):
            return False
        return self._field_name == other.field_name()

    def __hash__(self) -> int:
        return hash((self.arg_type(), tuple(self._field_name)))

    @staticmethod
    def builder() -> Builder:
        """The builder for creating a new instance of `FieldReferenceDTO`.

        Returns:
            Builder: The builder for creating a new instance of `FieldReferenceDTO`.
        """
        return FieldReferenceDTO.Builder()

    class Builder:
        """Builder for `FieldRererenceDTO`"""

        def __init__(self):
            self._field_name = None

        def with_field_name(self, field_name: List[str]) -> FieldReferenceDTO.Builder:
            """Set the field name for the field reference.

            Args:
                field_name (List[str]): The field name.

            Returns:
                FieldReferenceDTO.Builder: The builder.
            """

            self._field_name = field_name
            return self

        def with_column_name(self, column_name: str) -> FieldReferenceDTO.Builder:
            """Set the column name for the field reference.

            Args:
                column_name (str): The column name.

            Returns:
                FieldReferenceDTO.Builder: The builder.
            """

            self._field_name = [column_name]
            return self

        def build(self) -> FieldReferenceDTO:
            """Build the field reference.

            Returns:
                FieldReferenceDTO: The field reference.
            """

            return FieldReferenceDTO(field_name=self._field_name)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from dataclasses import dataclass, field
from typing import List, Optional, Union, cast

from dataclasses_json import DataClassJsonMixin, config

from gravitino.api.rel.column import Column
from gravitino.api.rel.expressions.expression import Expression
from gravitino.api.rel.types.json_serdes.type_serdes import TypeSerdes
from gravitino.api.rel.types.type import Type
from gravitino.api.rel.types.types import Types
from gravitino.dto.rel.expressions.json_serdes.column_default_value_serdes import (
    ColumnDefaultValueSerdes,
)
from gravitino.dto.rel.expressions.literal_dto import LiteralDTO
from gravitino.utils.precondition import Precondition


@dataclass
class ColumnDTO(Column, DataClassJsonMixin):
    """Represents a Model DTO (Data Transfer Object)."""

    _name: str = field(metadata=config(field_name="name"))
    """The name of the column."""

    _data_type: Type = field(
        metadata=config(
            field_name="type",
            encoder=TypeSerdes.serialize,
            decoder=TypeSerdes.deserialize,
        )
    )
    """The data type of the column."""

    _comment: Optional[str] = field(default=None, metadata=config(field_name="comment"))
    """The comment associated with the column."""

    _default_value: Optional[Union[Expression, List[Expression]]] = field(
        default_factory=lambda: Column.DEFAULT_VALUE_NOT_SET,
        metadata=config(
            field_name="defaultValue",
            encoder=ColumnDefaultValueSerdes.serialize,
            decoder=ColumnDefaultValueSerdes.deserialize,
            exclude=lambda value: value is None
            or value is Column.DEFAULT_VALUE_NOT_SET,
        ),
    )
    """The default value of the column."""

    _nullable: bool = field(default=True, metadata=config(field_name="nullable"))
    """Whether the column value can be null."""

    _auto_increment: bool = field(
        default=False, metadata=config(field_name="autoIncrement")
    )
    """Whether the column is an auto-increment column."""

    def name(self) -> str:
        return self._name

    def data_type(self) -> Type:
        return self._data_type

    def comment(self) -> str:
        return self._comment

    def nullable(self) -> bool:
        return self._nullable

    def auto_increment(self) -> bool:
        return self._auto_increment

    def default_value(self) -> Union[Expression, List[Expression]]:
        return self._default_value

    def validate(self) -> None:
        Precondition.check_string_not_empty(
            self._name, "Column name cannot be null or empty."
        )
        Precondition.check_argument(
            self._data_type is not None, "Column data type cannot be null."
        )
        non_nullable_condition = (
            not self._nullable
            and isinstance(self._default_value, LiteralDTO)
            and cast(LiteralDTO, self._default_value).data_type()
            == Types.NullType.get()
        )
        Precondition.check_argument(
            not non_nullable_condition,
            f"Column cannot be non-nullable with a null default value: {self._name}.",
        )

    @staticmethod
    def builder() -> ColumnDTO.Builder:
        return ColumnDTO.Builder()

    def __eq__(self, other: ColumnDTO) -> bool:
        if not isinstance(other, ColumnDTO):
            return False
        return (
            self._name == other._name
            and self._data_type == other._data_type
            and self._comment == other._comment
            and self._nullable == other._nullable
            and self._auto_increment == other._auto_increment
            and self._default_value == other._default_value
        )

    def __hash__(self) -> int:
        return hash(
            (
                self._name,
                self._data_type,
                self._comment,
                self._nullable,
                self._auto_increment,
                (
                    None
                    if self._default_value is Column.DEFAULT_VALUE_NOT_SET
                    else self._default_value
                ),
            )
        )

    class Builder:
        def __init__(self) -> None:
            self._name = None
            self._data_type = None
            self._comment = None
            self._nullable: bool = True
            self._auto_increment: bool = False
            self._default_value = None

        def with_name(self, name: str) -> ColumnDTO.Builder:
            """Sets the name of the column.

            Args:
                name (str): The name of the column.

            Returns:
                ColumnDTO.Builder: The builder instance.
            """
            self._name = name
            return self

        def with_data_type(self, data_type: Type) -> ColumnDTO.Builder:
            """Sets the data type of the column.

            Args:
                data_type (Type): The data type of the column.

            Returns:
                ColumnDTO.Builder: The builder instance.
            """
            self._data_type = data_type
            return self

        def with_comment(self, comment: str) -> ColumnDTO.Builder:
            """Sets the comment associated with the column.

            Args:
                comment (str): The comment associated with the column.

            Returns:
                ColumnDTO.Builder: The builder instance.
            """
            self._comment = comment
            return self

        def with_nullable(self, nullable: bool) -> ColumnDTO.Builder:
            """Sets whether the column value can be null.

            Args:
                nullable (bool): Whether the column value can be null.

            Returns:
                ColumnDTO.Builder: The builder instance.
            """
            self._nullable = nullable
            return self

        def with_auto_increment(self, auto_increment: bool) -> ColumnDTO.Builder:
            """Sets whether the column is an auto-increment column.

            Args:
                auto_increment (bool): Whether the column is an auto-increment column.

            Returns:
                ColumnDTO.Builder: The builder instance.
            """
            self._auto_increment = auto_increment
            return self

        def with_default_value(self, default_value: Expression) -> ColumnDTO.Builder:
            """Sets the default value of the column.

            Args:
                default_value (Expression): The default value of the column.

            Returns:
                ColumnDTO.Builder: The builder instance.
            """
            self._default_value = default_value
            return self

        def build(self) -> ColumnDTO:
            """Builds a Column DTO based on the provided builder parameters.

            Returns:
                ColumnDTO: A new ColumnDTO instance.

            Raises:
                IllegalArgumentException: If required, fields name and data type are not set.
            """
            Precondition.check_argument(
                self._name is not None, "Column name cannot be null"
            )
            Precondition.check_argument(
                self._data_type is not None, "Column data type cannot be null"
            )
            return ColumnDTO(
                _name=self._name,
                _data_type=self._data_type,
                _comment=self._comment,
                _nullable=self._nullable,
                _auto_increment=self._auto_increment,
                _default_value=(
                    Column.DEFAULT_VALUE_NOT_SET
                    if self._default_value is None
                    else self._default_value
                ),
            )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC
from typing import Dict

from gravitino.api.credential.credential import Credential
from gravitino.utils.precondition import Precondition


class S3TokenCredential(Credential, ABC):
    """Represents the S3 token credential."""

    S3_TOKEN_CREDENTIAL_TYPE: str = "s3-token"
    _SESSION_ACCESS_KEY_ID: str = "s3-access-key-id"
    _SESSION_SECRET_ACCESS_KEY: str = "s3-secret-access-key"
    _SESSION_TOKEN: str = "s3-session-token"

    _expire_time_in_ms: int = 0
    _access_key_id: str = None
    _secret_access_key: str = None
    _session_token: str = None

    def __init__(self, credential_info: Dict[str, str], expire_time_in_ms: int):
        self._access_key_id = credential_info.get(self._SESSION_ACCESS_KEY_ID, None)
        self._secret_access_key = credential_info.get(
            self._SESSION_SECRET_ACCESS_KEY, None
        )
        self._session_token = credential_info.get(self._SESSION_TOKEN, None)
        self._expire_time_in_ms = expire_time_in_ms
        Precondition.check_string_not_empty(
            self._access_key_id, "The S3 access key ID should not be empty"
        )
        Precondition.check_string_not_empty(
            self._secret_access_key, "The S3 secret access key should not be empty"
        )
        Precondition.check_string_not_empty(
            self._session_token, "The S3 session token should not be empty"
        )
        Precondition.check_argument(
            self._expire_time_in_ms > 0,
            "The expiration time of S3 token credential should be greater than 0",
        )

    def credential_type(self) -> str:
        """The type of the credential.

        Returns:
             the type of the credential.
        """
        return self.S3_TOKEN_CREDENTIAL_TYPE

    def expire_time_in_ms(self) -> int:
        """Returns the expiration time of the credential in milliseconds since
        the epoch, 0 means it will never expire.

        Returns:
             The expiration time of the credential.
        """
        return self._expire_time_in_ms

    def credential_info(self) -> Dict[str, str]:
        """The credential information.

        Returns:
             The credential information.
        """
        return {
            self._SESSION_ACCESS_KEY_ID: self._access_key_id,
            self._SESSION_SECRET_ACCESS_KEY: self._secret_access_key,
            self._SESSION_TOKEN: self._session_token,
        }

    def access_key_id(self) -> str:
        """The S3 access key id.

        Returns:
            The S3 access key id.
        """
        return self._access_key_id

    def secret_access_key(self) -> str:
        """The S3 secret access key.

        Returns:
            The S3 secret access key.
        """
        return self._secret_access_key

    def session_token(self) -> str:
        """The S3 session token.

        Returns:
            The S3 session token.
        """
        return self._session_token



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional

from dataclasses_json import config, dataclass_json

from gravitino.api.tag.tag import Tag
from gravitino.dto.audit_dto import AuditDTO


@dataclass_json
@dataclass
class TagDTO(Tag):
    """Represents a Tag Data Transfer Object (DTO)."""

    _name: str = field(metadata=config(field_name="name"))
    _comment: str = field(metadata=config(field_name="comment"))
    _properties: dict[str, str] = field(metadata=config(field_name="properties"))

    _audit: AuditDTO = field(default=None, metadata=config(field_name="audit"))
    _inherited: Optional[bool] = field(
        default=None, metadata=config(field_name="inherited")
    )

    def __eq__(self, other: object):
        if not isinstance(other, TagDTO):
            return False
        return (
            self._name == other._name
            and self._comment == other._comment
            and self._properties == other._properties
            and self._audit == other._audit
        )

    def __hash__(self) -> int:
        return hash(
            (
                self._name,
                self._comment,
                frozenset(self._properties.items()) if self._properties else None,
                self._audit,
            )
        )

    @staticmethod
    def builder() -> TagDTO.Builder:
        return TagDTO.Builder()

    def name(self) -> str:
        """Get the name of the tag.

        Returns:
            str: The name of the tag.
        """
        return self._name

    def comment(self) -> str:
        """Get the comment of the tag.

        Returns:
            str: The comment of the tag.
        """
        return self._comment

    def properties(self) -> dict[str, str]:
        """
        Get the properties of the tag.

        Returns:
            dict[str, str]: The properties of the tag.
        """
        return self._properties

    def audit_info(self) -> AuditDTO:
        """
        Get the audit information of the tag.

        Returns:
            AuditDTO: The audit information of the tag.
        """
        return self._audit

    def inherited(self) -> Optional[bool]:
        """Check if the tag is inherited from a parent object or not.

        If the tag is inherited, it will return `True`, if it is owned by the object itself, it will return `False`.

        **Note**. The return value is optional, only when the tag is associated with an object, and called from the
        object, the return value will be present. Otherwise, it will be empty.

        Returns:
            Optional[bool]:
                True if the tag is inherited, false if it is owned by the object itself. Empty if the
                tag is not associated with any object.
        """
        return self._inherited

    class Builder:
        """Helper class to build a TagDTO object."""

        def __init__(self) -> None:
            self._name = ""
            self._comment = ""
            self._properties: dict[str, str] = {}
            self._audit = None
            self._inherited = True

        def name(self, name: str) -> TagDTO.Builder:
            self._name = name
            return self

        def comment(self, comment: str) -> TagDTO.Builder:
            self._comment = comment
            return self

        def properties(self, properties: dict[str, str]) -> TagDTO.Builder:
            self._properties = properties
            return self

        def audit_info(self, audit: AuditDTO) -> TagDTO.Builder:
            self._audit = audit
            return self

        def inherited(self, inherited: bool) -> TagDTO.Builder:
            self._inherited = inherited
            return self

        def build(self) -> TagDTO:
            return TagDTO(
                self._name,
                self._comment,
                self._properties,
                self._audit,
                self._inherited,
            )
